# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `softmax.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions import softmax
from distrax._src.utils import equivalence
from distrax._src.utils import math
import jax
import jax.numpy as jnp
import numpy as np


class SoftmaxUnitTemperatureTest(equivalence.EquivalenceTest):

  def setUp(self):
    super().setUp()
    self._init_distr_cls(softmax.Softmax)
    self.temperature = 1.
    self.probs = jnp.array([0.2, 0.4, 0.1, 0.3])
    self.logits = jnp.log(self.probs)

  def test_num_categories(self):
    dist = self.distrax_cls(logits=self.logits)
    np.testing.assert_equal(dist.num_categories, len(self.logits))

  def test_parameters(self):
    dist = self.distrax_cls(logits=self.logits)
    self.assertion_fn(rtol=2e-3)(dist.logits, self.logits)
    self.assertion_fn(rtol=2e-3)(dist.probs, self.probs)


class SoftmaxTest(equivalence.EquivalenceTest):

  def setUp(self):
    super().setUp()
    self._init_distr_cls(softmax.Softmax)
    self.temperature = 10.
    self.logits = jnp.array([2., 4., 1., 3.])
    self.probs = jax.nn.softmax(self.logits / self.temperature)

  def test_num_categories(self):
    dist = self.distrax_cls(logits=self.logits, temperature=self.temperature)
    np.testing.assert_equal(dist.num_categories, len(self.logits))

  def test_parameters(self):
    dist = self.distrax_cls(logits=self.logits, temperature=self.temperature)
    self.assertion_fn(rtol=2e-3)(
        dist.logits, math.normalize(logits=self.logits / self.temperature))
    self.assertion_fn(rtol=2e-3)(dist.probs, self.probs)

  @chex.all_variants
  @parameterized.named_parameters(
      ('int32', jnp.int32),
      ('int64', jnp.int64),
      ('float32', jnp.float32),
      ('float64', jnp.float64))
  def test_sample_dtype(self, dtype):
    with jax.experimental.enable_x64(dtype.dtype.itemsize == 8):
      dist = self.distrax_cls(
          logits=self.logits, temperature=self.temperature, dtype=dtype)
      samples = self.variant(dist.sample)(seed=self.key)
      self.assertEqual(samples.dtype, dist.dtype)
      chex.assert_type(samples, dtype)

  def test_jittable(self):
    super()._test_jittable((np.array([2., 4., 1., 3.]),))

  @parameterized.named_parameters(
      ('single element', 2),
      ('range', slice(-1)),
      ('range_2', (slice(None), slice(-1))),
  )
  def test_slice(self, slice_):
    logits = jnp.array(np.random.randn(3, 4, 5))
    temperature = 0.8
    scaled_logits = logits / temperature
    dist = self.distrax_cls(logits=logits, temperature=temperature)
    self.assertIsInstance(dist[slice_], self.distrax_cls)
    self.assertion_fn(rtol=2e-3)(dist[slice_].temperature, temperature)
    self.assertion_fn(rtol=2e-3)(
        jax.nn.softmax(dist[slice_].logits, axis=-1),
        jax.nn.softmax(scaled_logits[slice_], axis=-1))

  def test_slice_ellipsis(self):
    logits = jnp.array(np.random.randn(3, 4, 5))
    temperature = 0.8
    scaled_logits = logits / temperature
    dist = self.distrax_cls(logits=logits, temperature=temperature)
    dist_sliced = dist[..., -1]
    self.assertIsInstance(dist_sliced, self.distrax_cls)
    self.assertion_fn(rtol=2e-3)(dist_sliced.temperature, temperature)
    self.assertion_fn(rtol=2e-3)(
        jax.nn.softmax(dist_sliced.logits, axis=-1),
        jax.nn.softmax(scaled_logits[:, -1], axis=-1))


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""MultivariateNormalTri distribution."""

from typing import Optional

import chex
from distrax._src.bijectors.diag_linear import DiagLinear
from distrax._src.bijectors.triangular_linear import TriangularLinear
from distrax._src.distributions import distribution
from distrax._src.distributions.mvn_from_bijector import MultivariateNormalFromBijector
from distrax._src.utils import conversion
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions

Array = chex.Array


def _check_parameters(
    loc: Optional[Array], scale_tri: Optional[Array]) -> None:
  """Checks that the inputs are correct."""

  if loc is None and scale_tri is None:
    raise ValueError(
        'At least one of `loc` and `scale_tri` must be specified.')

  if loc is not None and loc.ndim < 1:
    raise ValueError('The parameter `loc` must have at least one dimension.')

  if scale_tri is not None and scale_tri.ndim < 2:
    raise ValueError(
        f'The parameter `scale_tri` must have at least two dimensions, but '
        f'`scale_tri.shape = {scale_tri.shape}`.')

  if scale_tri is not None and scale_tri.shape[-1] != scale_tri.shape[-2]:
    raise ValueError(
        f'The parameter `scale_tri` must be a (batched) square matrix, but '
        f'`scale_tri.shape = {scale_tri.shape}`.')

  if loc is not None:
    num_dims = loc.shape[-1]
    if scale_tri is not None and scale_tri.shape[-1] != num_dims:
      raise ValueError(
          f'Shapes are not compatible: `loc.shape = {loc.shape}` and '
          f'`scale_tri.shape = {scale_tri.shape}`.')


class MultivariateNormalTri(MultivariateNormalFromBijector):
  """Multivariate normal distribution on `R^k`.

  The `MultivariateNormalTri` distribution is parameterized by a `k`-length
  location (mean) vector `b` and a (lower or upper) triangular scale matrix `S`
  of size `k x k`. The covariance matrix is `C = S @ S.T`.
  """

  equiv_tfp_cls = tfd.MultivariateNormalTriL

  def __init__(self,
               loc: Optional[Array] = None,
               scale_tri: Optional[Array] = None,
               is_lower: bool = True):
    """Initializes a MultivariateNormalTri distribution.

    Args:
      loc: Mean vector of the distribution of shape `k` (can also be a batch of
        such vectors). If not specified, it defaults to zeros.
      scale_tri: The scale matrix `S`. It must be a `k x k` triangular matrix
        (additional dimensions index batches). If `scale_tri` is not triangular,
        the entries above or below the main diagonal will be ignored. The
        parameter `is_lower` specifies if `scale_tri` is lower or upper
        triangular. It is the responsibility of the user to make sure that
        `scale_tri` only contains non-zero elements in its diagonal; this class
        makes no attempt to verify that. If `scale_tri` is not specified, it
        defaults to the identity.
      is_lower: Indicates if `scale_tri` is lower (if True) or upper (if False)
        triangular.
    """
    loc = None if loc is None else conversion.as_float_array(loc)
    scale_tri = None if scale_tri is None else conversion.as_float_array(
        scale_tri)
    _check_parameters(loc, scale_tri)

    num_dims = None
    if loc is not None:
      num_dims = loc.shape[-1]
    elif scale_tri is not None:
      num_dims = scale_tri.shape[-1]

    dtype = jnp.result_type(*[x for x in [loc, scale_tri] if x is not None])

    if loc is None:
      assert num_dims is not None
      loc = jnp.zeros((num_dims,), dtype=dtype)

    if scale_tri is None:
      self._scale_tri = jnp.eye(num_dims, dtype=dtype)
      scale = DiagLinear(diag=jnp.ones(loc.shape[-1:], dtype=dtype))
    else:
      tri_fn = jnp.tril if is_lower else jnp.triu
      self._scale_tri = tri_fn(scale_tri)
      scale = TriangularLinear(matrix=self._scale_tri, is_lower=is_lower)
    self._is_lower = is_lower

    super().__init__(loc=loc, scale=scale)

  @property
  def scale_tri(self) -> Array:
    """Triangular scale matrix `S`."""
    return jnp.broadcast_to(
        self._scale_tri,
        self.batch_shape + self.event_shape + self.event_shape)

  @property
  def is_lower(self) -> bool:
    """Whether the `scale_tri` matrix is lower triangular."""
    return self._is_lower

  def __getitem__(self, index) -> 'MultivariateNormalTri':
    """See `Distribution.__getitem__`."""
    index = distribution.to_batch_shape_index(self.batch_shape, index)
    return MultivariateNormalTri(
        loc=self.loc[index],
        scale_tri=self.scale_tri[index],
        is_lower=self.is_lower)



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Bijector abstract base class."""

import abc
import typing
from typing import Callable, Optional, Tuple, Union

import chex
from distrax._src.utils import jittable
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp


tfb = tfp.bijectors

Array = chex.Array


class Bijector(jittable.Jittable, metaclass=abc.ABCMeta):
  """Differentiable bijection that knows to compute its Jacobian determinant.

  A bijector implements a differentiable and bijective transformation `f`, whose
  inverse is also differentiable (`f` is called a "diffeomorphism"). A bijector
  can be used to transform a continuous random variable `X` to a continuous
  random variable `Y = f(X)` in the context of `TransformedDistribution`.

  Typically, a bijector subclass will implement the following methods:

  - `forward_and_log_det(x)` (required)
  - `inverse_and_log_det(y)` (optional)

  The remaining methods are defined in terms of the above by default.

  Subclass requirements:

  - Subclasses must ensure that `f` is differentiable and bijective, and that
    their methods correctly implement `f^{-1}`, `J(f)` and `J(f^{-1})`. Distrax
    will assume these properties hold, and will make no attempt to verify them.

  - Distrax assumes that `f` acts on array-valued variables called "events", and
    that the bijector operates on batched events. Specifically, Distrax assumes
    the following:

    * `f` acts on events of shape [M1, ..., Mn] and returns events of shape
      [L1, ..., Lq]. `n` is referred to as `event_ndims_in`, and `q` as
      `event_ndims_out`. `event_ndims_in` and `event_ndims_out` must be static
      properties of the bijector, and must be known to it at construction time.

    * The bijector acts on batched events of shape [N1, ..., Nk, M1, ..., Mn],
      where [N1, ..., Nk] are batch dimensions, and returns batched events of
      shape [K1, ..., Kp, L1, ..., Lq], where [K1, ..., Kp] are (possibly
      different) batch dimensions. Distrax requires that bijectors always
      broadcast against batched events, that is, that they apply `f` identically
      to each event. Distrax also allows for events to broadcast against batched
      bijectors, meaning that multiple instantiations of `f` are applied to the
      same event, although this is not a subclass requirement.
  """

  def __init__(self,
               event_ndims_in: int,
               event_ndims_out: Optional[int] = None,
               is_constant_jacobian: bool = False,
               is_constant_log_det: Optional[bool] = None):
    """Initializes a Bijector.

    Args:
      event_ndims_in: Number of input event dimensions. The bijector acts on
        events of shape [M1, ..., Mn], where `n == event_ndims_in`.
      event_ndims_out: Number of output event dimensions. The bijector returns
        events of shape [L1, ..., Lq], where `q == event_ndims_out`. If None, it
        defaults to `event_ndims_in`.
      is_constant_jacobian: Whether the Jacobian is promised to be constant
        (which is the case if and only if the bijector is affine). A value of
        False will be interpreted as "we don't know whether the Jacobian is
        constant", rather than "the Jacobian is definitely not constant". Only
        set to True if you're absolutely sure the Jacobian is constant; if
        you're not sure, set to False.
      is_constant_log_det: Whether the Jacobian determinant is promised to be
        constant (which is the case for, e.g., volume-preserving bijectors). If
        None, it defaults to `is_constant_jacobian`. Note that the Jacobian
        determinant can be constant without the Jacobian itself being constant.
        Only set to True if you're absoltely sure the Jacobian determinant is
        constant; if you're not sure, set to None.
    """
    if event_ndims_out is None:
      event_ndims_out = event_ndims_in
    if event_ndims_in < 0:
      raise ValueError(
          f"`event_ndims_in` can't be negative. Got {event_ndims_in}.")
    if event_ndims_out < 0:
      raise ValueError(
          f"`event_ndims_out` can't be negative. Got {event_ndims_out}.")
    if is_constant_log_det is None:
      is_constant_log_det = is_constant_jacobian
    if is_constant_jacobian and not is_constant_log_det:
      raise ValueError("The Jacobian is said to be constant, but its "
                       "determinant is said not to be, which is impossible.")
    self._event_ndims_in = event_ndims_in
    self._event_ndims_out = event_ndims_out
    self._is_constant_jacobian = is_constant_jacobian
    self._is_constant_log_det = is_constant_log_det

  def forward(self, x: Array) -> Array:
    """Computes y = f(x)."""
    y, _ = self.forward_and_log_det(x)
    return y

  def inverse(self, y: Array) -> Array:
    """Computes x = f^{-1}(y)."""
    x, _ = self.inverse_and_log_det(y)
    return x

  def forward_log_det_jacobian(self, x: Array) -> Array:
    """Computes log|det J(f)(x)|."""
    _, logdet = self.forward_and_log_det(x)
    return logdet

  def inverse_log_det_jacobian(self, y: Array) -> Array:
    """Computes log|det J(f^{-1})(y)|."""
    _, logdet = self.inverse_and_log_det(y)
    return logdet

  @abc.abstractmethod
  def forward_and_log_det(self, x: Array) -> Tuple[Array, Array]:
    """Computes y = f(x) and log|det J(f)(x)|."""

  def inverse_and_log_det(self, y: Array) -> Tuple[Array, Array]:
    """Computes x = f^{-1}(y) and log|det J(f^{-1})(y)|."""
    raise NotImplementedError(
        f"Bijector {self.name} does not implement `inverse_and_log_det`.")

  @property
  def event_ndims_in(self) -> int:
    """Number of input event dimensions."""
    return self._event_ndims_in

  @property
  def event_ndims_out(self) -> int:
    """Number of output event dimensions."""
    return self._event_ndims_out

  @property
  def is_constant_jacobian(self) -> bool:
    """Whether the Jacobian is promised to be constant."""
    return self._is_constant_jacobian

  @property
  def is_constant_log_det(self) -> bool:
    """Whether the Jacobian determinant is promised to be constant."""
    return self._is_constant_log_det

  @property
  def name(self) -> str:
    """Name of the bijector."""
    return self.__class__.__name__

  def same_as(self, other: "Bijector") -> bool:
    """Returns True if this bijector is guaranteed to be the same as `other`."""
    del other
    return False

  def _check_forward_input_shape(self, x: Array) -> None:
    """Checks that the input `x` to a forward method has valid shape."""
    x_ndims = len(jnp.shape(x))
    if x_ndims < self.event_ndims_in:
      raise ValueError(
          f"Bijector {self.name} has `event_ndims_in=={self.event_ndims_in}`,"
          f" but the input has only {x_ndims} array dimensions.")

  def _check_inverse_input_shape(self, y: Array) -> None:
    """Checks that the input `y` to an inverse method has valid shape."""
    y_ndims = len(jnp.shape(y))
    if y_ndims < self.event_ndims_out:
      raise ValueError(
          f"Bijector {self.name} has `event_ndims_out=={self.event_ndims_out}`,"
          f" but the input has only {y_ndims} array dimensions.")


BijectorLike = Union[Bijector, tfb.Bijector, Callable[[Array], Array]]
BijectorT = typing.TypeVar("BijectorT", bound=Bijector)



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""MultivariateNormalFullCovariance distribution."""

from typing import Optional

import chex
from distrax._src.distributions import distribution
from distrax._src.distributions.mvn_tri import MultivariateNormalTri
from distrax._src.utils import conversion
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions
Array = chex.Array


def _check_parameters(
    loc: Optional[Array], covariance_matrix: Optional[Array]) -> None:
  """Checks that the inputs are correct."""

  if loc is None and covariance_matrix is None:
    raise ValueError(
        'At least one of `loc` and `covariance_matrix` must be specified.')

  if loc is not None and loc.ndim < 1:
    raise ValueError('The parameter `loc` must have at least one dimension.')

  if covariance_matrix is not None and covariance_matrix.ndim < 2:
    raise ValueError(
        f'The `covariance_matrix` must have at least two dimensions, but '
        f'`covariance_matrix.shape = {covariance_matrix.shape}`.')

  if covariance_matrix is not None and (
      covariance_matrix.shape[-1] != covariance_matrix.shape[-2]):
    raise ValueError(
        f'The `covariance_matrix` must be a (batched) square matrix, but '
        f'`covariance_matrix.shape = {covariance_matrix.shape}`.')

  if loc is not None:
    num_dims = loc.shape[-1]
    if covariance_matrix is not None and (
        covariance_matrix.shape[-1] != num_dims):
      raise ValueError(
          f'Shapes are not compatible: `loc.shape = {loc.shape}` and '
          f'`covariance_matrix.shape = {covariance_matrix.shape}`.')


class MultivariateNormalFullCovariance(MultivariateNormalTri):
  """Multivariate normal distribution on `R^k`.

  The `MultivariateNormalFullCovariance` distribution is parameterized by a
  `k`-length location (mean) vector `b` and a covariance matrix `C` of size
  `k x k` that must be positive definite and symmetric.

  This class makes no attempt to verify that `C` is positive definite or
  symmetric. It is the responsibility of the user to make sure that it is the
  case.
  """

  equiv_tfp_cls = tfd.MultivariateNormalFullCovariance

  def __init__(self,
               loc: Optional[Array] = None,
               covariance_matrix: Optional[Array] = None):
    """Initializes a MultivariateNormalFullCovariance distribution.

    Args:
      loc: Mean vector of the distribution of shape `k` (can also be a batch of
        such vectors). If not specified, it defaults to zeros.
      covariance_matrix: The covariance matrix `C`. It must be a `k x k` matrix
        (additional dimensions index batches). If not specified, it defaults to
        the identity.
    """
    loc = None if loc is None else conversion.as_float_array(loc)
    covariance_matrix = None if covariance_matrix is None else (
        conversion.as_float_array(covariance_matrix))
    _check_parameters(loc, covariance_matrix)

    num_dims = None
    if loc is not None:
      num_dims = loc.shape[-1]
    elif covariance_matrix is not None:
      num_dims = covariance_matrix.shape[-1]

    dtype = jnp.result_type(
        *[x for x in [loc, covariance_matrix] if x is not None])

    if loc is None:
      assert num_dims is not None
      loc = jnp.zeros((num_dims,), dtype=dtype)

    if covariance_matrix is None:
      self._covariance_matrix = jnp.eye(num_dims, dtype=dtype)
      scale_tril = None
    else:
      self._covariance_matrix = covariance_matrix
      scale_tril = jnp.linalg.cholesky(covariance_matrix)

    super().__init__(loc=loc, scale_tri=scale_tril)

  @property
  def covariance_matrix(self) -> Array:
    """Covariance matrix `C`."""
    return jnp.broadcast_to(
        self._covariance_matrix,
        self.batch_shape + self.event_shape + self.event_shape)

  def covariance(self) -> Array:
    """Covariance matrix `C`."""
    return self.covariance_matrix

  def variance(self) -> Array:
    """Calculates the variance of all one-dimensional marginals."""
    return jnp.vectorize(jnp.diag, signature='(k,k)->(k)')(
        self.covariance_matrix)

  def __getitem__(self, index) -> 'MultivariateNormalFullCovariance':
    """See `Distribution.__getitem__`."""
    index = distribution.to_batch_shape_index(self.batch_shape, index)
    return MultivariateNormalFullCovariance(
        loc=self.loc[index],
        covariance_matrix=self.covariance_matrix[index])



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Hidden Markov Model implementation."""

from typing import Optional, Tuple

import chex
from distrax._src.distributions import categorical
from distrax._src.distributions import distribution
from distrax._src.utils import conversion
from distrax._src.utils import jittable
import jax
import jax.numpy as jnp


def _normalize(u: chex.Array,
               axis: int = 0,
               eps: float = 1e-15) -> Tuple[chex.Array, chex.Array]:
  """Normalizes the values within the axis in a way that they sum up to 1.

  Args:
    u: Input array to normalize.
    axis: Axis over which to normalize.
    eps: Minimum value threshold for numerical stability.

  Returns:
    Tuple of the normalized values, and the normalizing denominator.
  """
  u = jnp.where(u == 0, 0, jnp.where(u < eps, eps, u))
  c = u.sum(axis=axis)
  c = jnp.where(c == 0, 1, c)
  return u / c, c


class HMM(jittable.Jittable):
  """Hidden Markov Model class."""

  def __init__(self,
               init_dist: categorical.CategoricalLike,
               trans_dist: categorical.CategoricalLike,
               obs_dist: distribution.DistributionLike):
    """Constructs an N-state Hidden Markov Model from component distributions.

    Args:
      init_dist: Integer-valued categorical distribution with parameters of
        shape (N,), representing the distribution over initial latent states.
      trans_dist: Integer-valued categorical distribution with parameters of
        shape (N, N), representing the transition probability matrix between
        latent states.
      obs_dist: Any observation distribution with batch shape (N,), representing
        `p(observation|latent state)`.
    """
    self._init_dist = conversion.as_distribution(init_dist)
    self._trans_dist = conversion.as_distribution(trans_dist)
    self._obs_dist = conversion.as_distribution(obs_dist)
    self._n_states = self._init_dist.num_categories

    if not jnp.issubdtype(self._init_dist.dtype, jnp.integer):
      raise TypeError(
          f'init_dist must be categorical-like with integer dtype, but its '
          f'dtype is {self._init_dist.dtype}.')

    if not jnp.issubdtype(self._trans_dist.dtype, jnp.integer):
      raise TypeError(
          f'trans_dist must be categorical-like with integer dtype, but its '
          f'dtype is {self._trans_dist.dtype}.')

    if self._init_dist.batch_shape:
      raise ValueError(
          f'init_dist must be unbatched, but it has a batch shape of '
          f'{self._init_dist.batch_shape}.')

    if self._obs_dist.batch_shape != (self._n_states,):
      raise ValueError(
          f'obs_dist should have batch shape of ({self._n_states},) equal to '
          f'the number of latent states in the model, but its batch shape is '
          f'{self._obs_dist.batch_shape}.')

    if self._trans_dist.batch_shape != (self._n_states,):
      raise ValueError(
          f'trans_dist should have batch shape of ({self._n_states},) equal to '
          f'the number of latent states in the model, but its batch shape is '
          f'{self._trans_dist.batch_shape}.')

    if self._trans_dist.num_categories != self._n_states:
      raise ValueError(
          f'trans_dist should have `num_categories` of {self._n_states} equal '
          f'to the number of latent states in the model, but it has '
          f'`num_categories` of {self._trans_dist.num_categories}.')

  @property
  def init_dist(self) -> categorical.CategoricalLike:
    return self._init_dist

  @property
  def trans_dist(self) -> categorical.CategoricalLike:
    return self._trans_dist

  @property
  def obs_dist(self) -> distribution.DistributionLike:
    return self._obs_dist

  def sample(self,
             *,
             seed: chex.PRNGKey,
             seq_len: int) -> Tuple[chex.Array, chex.Array]:
    """Sample from this HMM.

    Samples an observation of given length according to this
    Hidden Markov Model and gives the sequence of the hidden states
    as well as the observation.

    Args:
      seed: Random key of shape (2,) and dtype uint32.
      seq_len: The length of the observation sequence.

    Returns:
      Tuple of hidden state sequence, and observation sequence.
    """
    rng_key, rng_init = jax.random.split(seed)
    initial_state = self._init_dist.sample(seed=rng_init)

    def draw_state(prev_state, key):
      state = self._trans_dist.sample(seed=key)[prev_state]
      return state, state

    rng_state, rng_obs = jax.random.split(rng_key)
    keys = jax.random.split(rng_state, seq_len - 1)
    _, states = jax.lax.scan(draw_state, initial_state, keys)
    states = jnp.append(initial_state, states)

    def draw_obs(state, key):
      return self._obs_dist.sample(seed=key)[state]

    keys = jax.random.split(rng_obs, seq_len)
    obs_seq = jax.vmap(draw_obs, in_axes=(0, 0))(states, keys)

    return states, obs_seq

  def forward(self,
              obs_seq: chex.Array,
              length: Optional[chex.Array] = None) -> Tuple[float, chex.Array]:
    """Calculates a belief state.

    Args:
      obs_seq: Observation sequence.
      length: The valid length of the observation sequence, used to truncate the
        computation for batches of varying length. If set to None, the entire
        sequence is used.

    Returns:
      Tuple of `log(p(x_{1:T}|model))` and the array of forward joint
        probabilities `p(z_t,x_{1:t})` for each sample `x_t`.
    """
    seq_len = len(obs_seq)

    if length is None:
      length = seq_len

    def scan_fn(carry, t):
      (alpha_prev, log_ll_prev) = carry
      alpha_n = jnp.where(
          t < length,
          (self._obs_dist.prob(obs_seq[t])
           * (alpha_prev[:, None] * self._trans_dist.probs).sum(axis=0)),
          jnp.zeros_like(alpha_prev))

      alpha_n, cn = _normalize(alpha_n)
      carry = (alpha_n, jnp.log(cn) + log_ll_prev)

      return carry, alpha_n

    # initial belief state
    alpha_0, c0 = _normalize(
        self._init_dist.probs * self._obs_dist.prob(obs_seq[0]))

    # setup scan loop
    init_state = (alpha_0, jnp.log(c0))
    ts = jnp.arange(1, seq_len)
    carry, alpha_hist = jax.lax.scan(scan_fn, init_state, ts)

    # post-process
    alpha_hist = jnp.vstack([alpha_0.reshape(1, self._n_states), alpha_hist])
    (_, log_ll) = carry
    return log_ll, alpha_hist

  def backward(self,
               obs_seq: chex.Array,
               length: Optional[chex.Array] = None) -> chex.Array:
    """Computes the backward probabilities.

    Args:
      obs_seq: Observation sequence.
      length: The valid length of the observation sequence, used to truncate the
        computation for batches of varying length. If set to None, the entire
        sequence is used.

    Returns:
      Array of backward joint probabilities `p(x_{t+1:T}|z_t)`.
    """
    seq_len = len(obs_seq)

    if length is None:
      length = seq_len

    beta_t = jnp.ones((self._n_states,))

    def scan_fn(beta_prev, t):
      beta_t = jnp.where(
          t > length,
          jnp.zeros_like(beta_prev),
          _normalize((beta_prev * self._obs_dist.prob(obs_seq[t-1])
                      * self._trans_dist.probs).sum(axis=1))[0])
      return beta_t, beta_t

    ts = jnp.arange(seq_len, 1, -1)
    _, beta_hist = jax.lax.scan(scan_fn, beta_t, ts)

    beta_hist = jnp.flip(
        jnp.vstack([beta_t.reshape(1, self._n_states), beta_hist]), axis=0)

    return beta_hist

  def forward_backward(
      self,
      obs_seq: chex.Array,
      length: Optional[chex.Array] = None,
  ) -> Tuple[chex.Array, chex.Array, chex.Array, float]:
    """HMM forward-backward algorithm.

    Computes, for each time step, the marginal conditional probability that the
    Hidden Markov Model was in each possible state given the observations that
    were made at each time step, i.e. P(z[i] | x[0], ..., x[num_steps - 1])
    for all i from 0 to num_steps - 1.

    Args:
      obs_seq: Observation sequence.
      length: The valid length of the observation sequence, used to truncate the
        computation for batches of varying length. If set to None, the entire
        sequence is used.

    Returns:
      Tuple of:
        * Forward joint probabilities `p(z_t,x_{1:t})`.
        * Backward joint probabilities `p(x_{t+1:T}|z_t)`.
        * Marginal conditional probability of the observations.
        * The log-likelihood log(p(x_{1:T}|model)).
    """
    seq_len = len(obs_seq)

    if length is None:
      length = seq_len

    def gamma_t(t):
      return alpha[t] * beta[t]

    ll, alpha = self.forward(obs_seq, length)

    beta = self.backward(obs_seq, length)

    ts = jnp.arange(seq_len)
    gamma = jax.vmap(gamma_t)(ts)
    gamma = jax.vmap(lambda x: _normalize(x)[0])(gamma)
    return alpha, beta, gamma, ll

  def viterbi(self, obs_seq: chex.Array) -> chex.Array:
    """Viterbi algorithm.

    Computes the most probable sequence of hidden states given the observations.

    Args:
      obs_seq: Observation sequence.

    Returns:
      The most probable sequence of hidden states.
    """

    trans_log_probs = jax.nn.log_softmax(self._trans_dist.logits)
    init_log_probs = jax.nn.log_softmax(self._init_dist.logits)

    first_log_prob = init_log_probs + self._obs_dist.log_prob(obs_seq[0])

    if len(obs_seq) == 1:
      return jnp.expand_dims(jnp.argmax(first_log_prob), axis=0)

    def viterbi_forward(prev_logp, obs):
      obs_logp = self._obs_dist.log_prob(obs)
      logp = prev_logp[..., None] + trans_log_probs + obs_logp[..., None, :]
      max_logp_given_successor = jnp.max(logp, axis=-2)
      most_likely_given_successor = jnp.argmax(logp, axis=-2)
      return max_logp_given_successor, most_likely_given_successor

    final_log_prob, most_likely_sources = jax.lax.scan(
        viterbi_forward, first_log_prob, obs_seq[1:])

    most_likely_initial_given_successor = jnp.argmax(
        trans_log_probs + first_log_prob, axis=-2)
    most_likely_sources = jnp.concatenate([
        jnp.expand_dims(most_likely_initial_given_successor, axis=0),
        most_likely_sources], axis=0)

    def viterbi_backward(state, most_likely_sources):
      state = jax.nn.one_hot(state, self._n_states)
      most_likely = jnp.sum(most_likely_sources * state).astype(jnp.int64)
      return most_likely, most_likely

    final_state = jnp.argmax(final_log_prob)
    _, most_likely_path = jax.lax.scan(
        viterbi_backward, final_state, most_likely_sources[1:], reverse=True)

    return jnp.append(most_likely_path, final_state)



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Diagonal-plus-low-rank linear bijector."""

from typing import Tuple

from distrax._src.bijectors import bijector as base
from distrax._src.bijectors import chain
from distrax._src.bijectors import diag_linear
from distrax._src.bijectors import linear
import jax
import jax.numpy as jnp

Array = base.Array


def _get_small_matrix(u_matrix: Array, v_matrix: Array) -> Array:
  rank = u_matrix.shape[-1]
  return jnp.eye(rank) + v_matrix.T @ u_matrix


def _get_logdet(matrix: Array) -> Array:
  """Computes the log absolute determinant of `matrix`."""
  return jnp.linalg.slogdet(matrix)[1]


def _forward_unbatched(x: Array, u_matrix: Array, v_matrix: Array) -> Array:
  return x + u_matrix @ (v_matrix.T @ x)


def _inverse_unbatched(
    y: Array, u_matrix: Array, v_matrix: Array, small_matrix: Array) -> Array:
  return y - u_matrix @ jax.scipy.linalg.solve(small_matrix, v_matrix.T @ y)


class _IdentityPlusLowRankLinear(base.Bijector):
  """Linear bijector whose weights are a low-rank perturbation of the identity.

  The bijector is defined as `f(x) = Ax` where `A = I + UV^T` and `U`, `V` are
  DxK matrices. When K < D, this bijector is computationally more efficient than
  an equivalent `UnconstrainedAffine` bijector.

  The Jacobian determinant is computed using the matrix determinant lemma:

  det J(x) = det A = det(I + V^T U)

  The matrix `I + V^T U` is KxK instead of DxD, so for K < D computing its
  determinant is faster than computing the determinant of `A`.

  The inverse is computed using the Woodbury matrix identity:

  A^{-1} = I - U (I + V^T U)^{-1} V^T

  As above, inverting the KxK matrix `I + V^T U` is faster than inverting `A`
  when K < D.

  The bijector is invertible if and only if `I + V^T U` is invertible. It is the
  responsibility of the user to make sure that this is the case; the class will
  make no attempt to verify that the bijector is invertible.
  """

  def __init__(self, u_matrix: Array, v_matrix: Array):
    """Initializes the bijector.

    Args:
      u_matrix: a DxK matrix, the `U` matrix in `A = I + UV^T`. Can also be a
        batch of DxK matrices.
      v_matrix: a DxK matrix, the `V` matrix in `A = I + UV^T`. Can also be a
        batch of DxK matrices.
    """
    super().__init__(event_ndims_in=1, is_constant_jacobian=True)
    self._batch_shape = jax.lax.broadcast_shapes(
        u_matrix.shape[:-2], v_matrix.shape[:-2])
    self._u_matrix = u_matrix
    self._v_matrix = v_matrix
    self._small_matrix = jnp.vectorize(
        _get_small_matrix, signature="(d,k),(d,k)->(k,k)")(u_matrix, v_matrix)
    self._logdet = _get_logdet(self._small_matrix)

  def forward(self, x: Array) -> Array:
    """Computes y = f(x)."""
    self._check_forward_input_shape(x)
    batched = jnp.vectorize(
        _forward_unbatched, signature="(d),(d,k),(d,k)->(d)")
    return batched(x, self._u_matrix, self._v_matrix)

  def forward_log_det_jacobian(self, x: Array) -> Array:
    """Computes log|det J(f)(x)|."""
    self._check_forward_input_shape(x)
    batch_shape = jax.lax.broadcast_shapes(self._batch_shape, x.shape[:-1])
    return jnp.broadcast_to(self._logdet, batch_shape)

  def forward_and_log_det(self, x: Array) -> Tuple[Array, Array]:
    """Computes y = f(x) and log|det J(f)(x)|."""
    return self.forward(x), self.forward_log_det_jacobian(x)

  def inverse(self, y: Array) -> Array:
    """Computes x = f^{-1}(y)."""
    self._check_inverse_input_shape(y)
    batched = jnp.vectorize(
        _inverse_unbatched, signature="(d),(d,k),(d,k),(k,k)->(d)")
    return batched(y, self._u_matrix, self._v_matrix, self._small_matrix)

  def inverse_log_det_jacobian(self, y: Array) -> Array:
    """Computes log|det J(f^{-1})(y)|."""
    return -self.forward_log_det_jacobian(y)

  def inverse_and_log_det(self, y: Array) -> Tuple[Array, Array]:
    """Computes x = f^{-1}(y) and log|det J(f^{-1})(y)|."""
    return self.inverse(y), self.inverse_log_det_jacobian(y)


def _check_shapes_are_valid(diag: Array,
                            u_matrix: Array,
                            v_matrix: Array) -> None:
  """Checks array shapes are valid, raises `ValueError` if not."""
  for x, name, n in [(diag, "diag", 1),
                     (u_matrix, "u_matrix", 2),
                     (v_matrix, "v_matrix", 2)]:
    if x.ndim < n:
      raise ValueError(
          f"`{name}` must have at least {n} dimensions, got {x.ndim}.")
  dim = diag.shape[-1]
  u_shape = u_matrix.shape[-2:]
  v_shape = v_matrix.shape[-2:]
  if u_shape[0] != dim:
    raise ValueError(
        f"The length of `diag` must equal the first dimension of `u_matrix`. "
        f"Got `diag.length = {dim}` and `u_matrix.shape = {u_shape}`.")
  if u_shape != v_shape:
    raise ValueError(
        f"`u_matrix` and `v_matrix` must have the same shape; got "
        f"`u_matrix.shape = {u_shape}` and `v_matrix.shape = {v_shape}`.")


class DiagPlusLowRankLinear(linear.Linear):
  """Linear bijector whose weights are a low-rank perturbation of a diagonal.

  The bijector is defined as `f(x) = Ax` where `A = S + UV^T` and:
  - `S` is a DxD diagonal matrix,
  - `U`, `V` are DxK matrices.
  When K < D, this bijector is computationally more efficient than an equivalent
  `UnconstrainedAffine` bijector.

  The Jacobian determinant is computed using the matrix determinant lemma:

  det J(x) = det A = det(S) det(I + V^T S^{-1} U)

  The matrix `I + V^T S^{-1} U` is KxK instead of DxD, so for K < D computing
  its determinant is faster than computing the determinant of `A`.

  The inverse is computed using the Woodbury matrix identity:

  A^{-1} = (I - S^{-1} U (I + V^T S^{-1} U)^{-1} V^T) S^{-1}

  As above, inverting the KxK matrix `I + V^T S^{-1} U` is faster than inverting
  `A` when K < D.

  The bijector is invertible if and only if both `S` and `I + V^T S^{-1} U` are
  invertible matrices. It is the responsibility of the user to make sure that
  this is the case; the class will make no attempt to verify that the bijector
  is invertible.
  """

  def __init__(self, diag: Array, u_matrix: Array, v_matrix: Array):
    """Initializes the bijector.

    Args:
      diag: a vector of length D, the diagonal of matrix `S`. Can also be a
        batch of such vectors.
      u_matrix: a DxK matrix, the `U` matrix in `A = S + UV^T`. Can also be a
        batch of DxK matrices.
      v_matrix: a DxK matrix, the `V` matrix in `A = S + UV^T`. Can also be a
        batch of DxK matrices.
    """
    _check_shapes_are_valid(diag, u_matrix, v_matrix)
    # Since `S + UV^T = S (I + WV^T)` where `W = S^{-1}U`, we can implement this
    # bijector by composing `_IdentityPlusLowRankLinear` with `DiagLinear`.
    id_plus_low_rank_linear = _IdentityPlusLowRankLinear(
        u_matrix=u_matrix / diag[..., None],
        v_matrix=v_matrix)
    self._bijector = chain.Chain(
        [diag_linear.DiagLinear(diag), id_plus_low_rank_linear])
    batch_shape = jnp.broadcast_shapes(
        diag.shape[:-1], u_matrix.shape[:-2], v_matrix.shape[:-2])
    dtype = jnp.result_type(diag, u_matrix, v_matrix)
    super().__init__(
        event_dims=diag.shape[-1], batch_shape=batch_shape, dtype=dtype)
    self._diag = diag
    self._u_matrix = u_matrix
    self._v_matrix = v_matrix
    self.forward = self._bijector.forward
    self.forward_log_det_jacobian = self._bijector.forward_log_det_jacobian
    self.inverse = self._bijector.inverse
    self.inverse_log_det_jacobian = self._bijector.inverse_log_det_jacobian
    self.inverse_and_log_det = self._bijector.inverse_and_log_det

  @property
  def diag(self) -> Array:
    """Vector of length D, the diagonal of matrix `S`."""
    return self._diag

  @property
  def u_matrix(self) -> Array:
    """The `U` matrix in `A = S + UV^T`."""
    return self._u_matrix

  @property
  def v_matrix(self) -> Array:
    """The `V` matrix in `A = S + UV^T`."""
    return self._v_matrix

  @property
  def matrix(self) -> Array:
    """The matrix `A = S + UV^T` of the transformation."""
    batched = jnp.vectorize(
        lambda s, u, v: jnp.diag(s) + u @ v.T,
        signature="(d),(d,k),(d,k)->(d,d)")
    return batched(self._diag, self._u_matrix, self._v_matrix)

  def forward_and_log_det(self, x: Array) -> Tuple[Array, Array]:
    """Computes y = f(x) and log|det J(f)(x)|."""
    return self._bijector.forward_and_log_det(x)

  def same_as(self, other: base.Bijector) -> bool:
    """Returns True if this bijector is guaranteed to be the same as `other`."""
    if type(other) is DiagPlusLowRankLinear:  # pylint: disable=unidiomatic-typecheck
      return all((
          self.diag is other.diag,
          self.u_matrix is other.u_matrix,
          self.v_matrix is other.v_matrix,
      ))
    return False



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utility functions for testing equivalence between TFP and Distrax."""

import functools
from typing import Any, Callable, Dict, Optional, Tuple, Union

from absl.testing import parameterized

import chex
from distrax._src.distributions import distribution
import jax
import numpy as np
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions

Array = chex.Array


def get_tfp_equiv(distrax_cls):
  """Returns the TFP equivalent of a Distrax class.

  Args:
    distrax_cls: The Distrax class or the name of the class as a string.

  Returns:
    The equivalent TFP class if found, else `None`.

  Raises:
    ValueError:
      If `distrax_cls` is neither a Distrax class nor a string.

  #### Examples
  from distrax import normal
  from distrax import tfp_utils

  # equivalent to tfd.normal.Normal(0, 1)
  tfp_utils.get_tfp_equiv(normal.Normal)(0, 1)

  # equivalent to tfd.normal.Normal(0, 1)
  tfp_utils.get_tfp_equiv("Normal")(0, 1)

  """

  if isinstance(distrax_cls, str):
    tfp_class_str = distrax_cls
  elif issubclass(distrax_cls, distribution.Distribution):
    tfp_class_str = distrax_cls.__name__
  else:
    raise ValueError(
        'distrax_cls must be the class object or the name of the class object'
        ' as a string'
    )

  if hasattr(tfd, tfp_class_str):
    return getattr(tfd, tfp_class_str)
  else:
    return None


class EquivalenceTest(parameterized.TestCase):
  """Provides comparison assertions for TFP and Distrax distributions."""

  def setUp(self):
    super().setUp()
    self.tfp_cls = None

  def _init_distr_cls(self, distrax_cls: type(distribution.Distribution)):
    self.key = jax.random.PRNGKey(1234)
    self.distrax_cls = distrax_cls
    if hasattr(distrax_cls, 'equiv_tfp_cls'):
      self.tfp_cls = distrax_cls.equiv_tfp_cls
    else:
      self.tfp_cls = get_tfp_equiv(distrax_cls)

  def assertion_fn(self, **kwargs) -> Callable[[Any, Any], None]:
    def f(x, y):
      np.testing.assert_allclose(x, y, **kwargs)

    return f

  def _get_tfp_cls(self) -> type(tfd.Distribution):
    if self.tfp_cls is None:
      raise ValueError('TFP class undefined. Run _init_distr_cls() first.')
    return self.tfp_cls

  def _test_attribute(
      self,
      attribute_string: str,
      dist_args: Tuple[Any, ...] = (),
      dist_kwargs: Optional[Dict[str, Any]] = None,
      tfp_dist_args: Optional[Tuple[Any, ...]] = None,
      tfp_dist_kwargs: Optional[Dict[str, Any]] = None,
      call_args: Tuple[Any, ...] = (),
      call_kwargs: Optional[Dict[str, Any]] = None,
      assertion_fn: Callable[[Any, Any], None] = np.testing.assert_allclose):
    """Asserts equivalence of TFP and Distrax attributes.

    Given a potentially callable attribute as a string, compares the attribute
    values among Distrax and TFP implementations.

    Args:
      attribute_string: An attribute or a method of a Distrax/TFP
        distribution, provided as a string.
      dist_args: Arguments to be passed to Distrax constructor as *dist_args.
      dist_kwargs: Keyword arguments to be passed to Distrax constructor as
        **dist_kwargs.
      tfp_dist_args: Arguments to be passed to TFP constructor as
        *tfp_dist_args. If None, defaults to `dist_args`.
      tfp_dist_kwargs: Keyword arguments to be passed to TFP constructor as
        **tfp_dist_kwargs. If None, defaults to `dist_kwargs`.
      call_args: Arguments to be passed to Distrax and TFP methods as
        *call_args.
      call_kwargs: Keyword arguments to be passed to Distrax and TFP
        methods as **call_kwargs.
      assertion_fn: Assertion function to be called to compare
        Distrax and TFP methods/attributes.
    """

    if dist_kwargs is None:
      dist_kwargs = {}
    if call_kwargs is None:
      call_kwargs = {}
    if tfp_dist_args is None:
      tfp_dist_args = dist_args
    if tfp_dist_kwargs is None:
      tfp_dist_kwargs = dist_kwargs

    dist = self.distrax_cls(*dist_args, **dist_kwargs)
    tfp_dist = self._get_tfp_cls()(*tfp_dist_args, **tfp_dist_kwargs)

    if callable(getattr(dist, attribute_string)):
      distrax_fn = getattr(dist, attribute_string)
      tfp_fn = getattr(tfp_dist, attribute_string)
      if hasattr(self, 'variant'):
        distrax_fn = self.variant(distrax_fn)
      assertion_fn(distrax_fn(*call_args, **call_kwargs),
                   tfp_fn(*call_args, **call_kwargs))
    else:
      assertion_fn(getattr(dist, attribute_string),
                   getattr(tfp_dist, attribute_string))

  def _test_event_shape(self, dist_args, dist_kwargs,
                        tfp_dist_args=None, tfp_dist_kwargs=None):
    """Tests event shape."""
    self._test_attribute('event_shape', dist_args, dist_kwargs,
                         tfp_dist_args, tfp_dist_kwargs)

  def _test_batch_shape(self, dist_args, dist_kwargs,
                        tfp_dist_args=None, tfp_dist_kwargs=None):
    """Tests batch shape."""
    self._test_attribute('batch_shape', dist_args, dist_kwargs,
                         tfp_dist_args, tfp_dist_kwargs)

  def _test_prob(self, dist_args, dist_kwargs, value,
                 tfp_dist_args=None, tfp_dist_kwargs=None):
    """Tests prob."""
    self._test_attribute('prob', dist_args, dist_kwargs, tfp_dist_args,
                         tfp_dist_kwargs, (value,))

  def _test_log_prob(self, dist_args, dist_kwargs, value,
                     tfp_dist_args=None, tfp_dist_kwargs=None):
    """Tests log prob."""
    assertion_fn = functools.partial(np.testing.assert_allclose, rtol=1e-2)
    self._test_attribute(
        'log_prob', dist_args, dist_kwargs, tfp_dist_args, tfp_dist_kwargs,
        (value,), assertion_fn=assertion_fn)

  def _test_cdf(self, dist_args, dist_kwargs, value,
                tfp_dist_args=None, tfp_dist_kwargs=None):
    """Tests CDF."""
    self._test_attribute('cdf', dist_args, dist_kwargs,
                         tfp_dist_args, tfp_dist_kwargs, (value,))

  def _test_log_cdf(self, dist_args, dist_kwargs, value,
                    tfp_dist_args=None, tfp_dist_kwargs=None):
    """Tests log CDF."""
    self._test_attribute('log_cdf', dist_args, dist_kwargs,
                         tfp_dist_args, tfp_dist_kwargs, (value,))

  def _test_sample_shape(self, dist_args, dist_kwargs, sample_shape,
                         tfp_dist_args=None, tfp_dist_kwargs=None):
    """Tests sample shape."""
    if tfp_dist_args is None:
      tfp_dist_args = dist_args
    if tfp_dist_kwargs is None:
      tfp_dist_kwargs = dist_kwargs
    dist = self.distrax_cls(*dist_args, **dist_kwargs)

    def sample_fn(key, sample_shape=sample_shape):
      return dist.sample(seed=key, sample_shape=sample_shape)

    if hasattr(self, 'variant'):
      sample_fn = self.variant(sample_fn)
    samples = sample_fn(self.key)

    tfp_dist = self._get_tfp_cls()(*tfp_dist_args, **tfp_dist_kwargs)
    tfp_samples = tfp_dist.sample(sample_shape=sample_shape,
                                  seed=self.key)
    chex.assert_equal_shape([samples, tfp_samples])

  def _test_sample_and_log_prob(
      self,
      dist_args: Tuple[Any, ...] = (),
      dist_kwargs: Optional[Dict[str, Any]] = None,
      tfp_dist_args: Optional[Tuple[Any, ...]] = None,
      tfp_dist_kwargs: Optional[Dict[str, Any]] = None,
      sample_shape: Union[int, Tuple[int, ...]] = (),
      assertion_fn: Callable[[Any, Any], None] = np.testing.assert_allclose):
    """Tests sample and log prob."""
    if tfp_dist_args is None:
      tfp_dist_args = dist_args
    if tfp_dist_kwargs is None:
      tfp_dist_kwargs = dist_kwargs
    dist = self.distrax_cls(*dist_args, **dist_kwargs)
    log_prob_fn = dist.log_prob

    def sample_and_log_prob_fn(key):
      return dist.sample_and_log_prob(seed=key, sample_shape=sample_shape)

    if hasattr(self, 'variant'):
      sample_and_log_prob_fn = self.variant(sample_and_log_prob_fn)
      log_prob_fn = self.variant(dist.log_prob)
    samples, log_prob = sample_and_log_prob_fn(self.key)

    tfp_dist = self._get_tfp_cls()(*tfp_dist_args, **tfp_dist_kwargs)
    tfp_samples = tfp_dist.sample(sample_shape=sample_shape,
                                  seed=self.key)
    tfp_log_prob = tfp_dist.log_prob(samples)

    chex.assert_equal_shape([samples, tfp_samples])
    assertion_fn(log_prob, tfp_log_prob)
    assertion_fn(log_prob, log_prob_fn(samples))

  def _test_with_two_distributions(
      self,
      attribute_string: str,
      mode_string: str = 'distrax_to_distrax',
      dist1_args: Tuple[Any, ...] = (),
      dist1_kwargs: Optional[Dict[str, Any]] = None,
      dist2_args: Tuple[Any, ...] = (),
      dist2_kwargs: Optional[Dict[str, Any]] = None,
      tfp_dist1_args: Tuple[Any, ...] = (),
      tfp_dist1_kwargs: Optional[Dict[str, Any]] = None,
      tfp_dist2_args: Tuple[Any, ...] = (),
      tfp_dist2_kwargs: Optional[Dict[str, Any]] = None,
      assertion_fn: Callable[[Any, Any], None] = np.testing.assert_allclose):
    """Asserts equivalence of TFP and Distrax methods that compare two distribs.

    This checks that the methods `D(dist1 || dist2)` and `D(dist2 || dist1)`
    give the same results as their TFP counterparts, where `D` is typically the
    KL divergence or the cross-entropy.

    Args:
      attribute_string: The method attribute, provided as a string.
      mode_string: string, must be one of the following:
        - If "distrax_to_distrax", this method verifies the values of
          `D(dist1 || dist2)` and `D(dist2 || dist1)`, where both `dist1` and
          `dist2` are Distrax distributions.
        - If "distrax_to_tfp", this method verifies the values of
          `D(dist1 || tfp_dist2)` and `D(dist2 || tfp_dist1)`.
        - If "tfp_to_distrax", this method verifies the values of
          `D(tfp_dist1 || dist2)` and `D(tfp_dist2 || dist1)`.
      dist1_args: Arguments to be passed to Distrax constructor as *dist_args
        for the first distribution
      dist1_kwargs: Keyword arguments to be passed to Distrax constructor as
        **dist_kwargs for the first distribution.
      dist2_args: Arguments to be passed to Distrax constructor as *dist_args
        for the second distribution.
      dist2_kwargs: Keyword arguments to be passed to Distrax constructor as
        **dist_kwargs for the second distribution.
      tfp_dist1_args: Arguments to be passed to TFP constructor as
        *tfp_dist_args for the first distribution. If None, defaults to
        `dist1_args`.
      tfp_dist1_kwargs: Keyword arguments to be passed to TFP constructor as
        **tfp_dist_kwargs for the first distribution. If None, defaults to
        `dist1_kwargs`.
      tfp_dist2_args: Arguments to be passed to TFP constructor as
        *tfp_dist_args for the second distribution. If None, defaults to
        `dist2_args`.
      tfp_dist2_kwargs: Keyword arguments to be passed to TFP constructor as
        **tfp_dist_kwargs for the second distribution. If None, defaults to
        `dist2_kwargs`.
      assertion_fn: Assertion function to be called to compare Distrax and TFP
        function values.
    """
    dist1_kwargs = {} if dist1_kwargs is None else dist1_kwargs
    dist2_kwargs = {} if dist2_kwargs is None else dist2_kwargs

    if tfp_dist1_args is None:
      tfp_dist1_args = dist1_args
    if tfp_dist1_kwargs is None:
      tfp_dist1_kwargs = dist1_kwargs
    if tfp_dist2_args is None:
      tfp_dist2_args = dist2_args
    if tfp_dist2_kwargs is None:
      tfp_dist2_kwargs = dist2_kwargs

    dist1 = self.distrax_cls(*dist1_args, **dist1_kwargs)
    tfp_dist1 = self._get_tfp_cls()(*tfp_dist1_args, **tfp_dist1_kwargs)
    dist2 = self.distrax_cls(*dist2_args, **dist2_kwargs)
    tfp_dist2 = self._get_tfp_cls()(*tfp_dist2_args, **tfp_dist2_kwargs)

    tfp_comp_dist1_dist2 = getattr(tfp_dist1, attribute_string)(tfp_dist2)
    tfp_comp_dist2_dist1 = getattr(tfp_dist2, attribute_string)(tfp_dist1)

    distrax_fn_1 = getattr(dist1, attribute_string)
    distrax_fn_2 = getattr(dist2, attribute_string)
    if hasattr(self, 'variant'):
      distrax_fn_1 = self.variant(distrax_fn_1)
      distrax_fn_2 = self.variant(distrax_fn_2)

    if mode_string == 'distrax_to_distrax':
      comp_dist1_dist2 = distrax_fn_1(dist2)
      comp_dist2_dist1 = distrax_fn_2(dist1)
    elif mode_string == 'distrax_to_tfp':
      comp_dist1_dist2 = distrax_fn_1(tfp_dist2)
      comp_dist2_dist1 = distrax_fn_2(tfp_dist1)
    elif mode_string == 'tfp_to_distrax':
      comp_dist1_dist2 = getattr(tfp_dist1, attribute_string)(dist2)
      comp_dist2_dist1 = getattr(tfp_dist2, attribute_string)(dist1)
    else:
      raise ValueError(
          f'`mode_string` should be one of the following: '
          f'"distrax_to_distrax", "distrax_to_tfp", or "tfp_to_distrax", '
          f'but it is "{mode_string}".')

    assertion_fn(comp_dist1_dist2, tfp_comp_dist1_dist2)
    assertion_fn(comp_dist2_dist1, tfp_comp_dist2_dist1)

  def _test_jittable(
      self,
      dist_args: Tuple[Any, ...] = (),
      dist_kwargs: Optional[Dict[str, Any]] = None,
      assertion_fn: Callable[[Any, Any], None] = np.testing.assert_allclose):
    """Tests that the distribution can be passed to a jitted function."""
    dist_kwargs = dist_kwargs or {}

    @jax.jit
    def jitted_function(event, dist):
      return dist.log_prob(event)

    dist = self.distrax_cls(*dist_args, **dist_kwargs)
    event = dist.sample(seed=self.key)
    log_prob = dist.log_prob(event)
    jitted_log_prob = jitted_function(event, dist)
    assertion_fn(jitted_log_prob, log_prob)

  def _test_raises_error(
      self,
      dist_args: Tuple[Any, ...] = (),
      dist_kwargs: Optional[Dict[str, Any]] = None,
      error_type=AssertionError):
    """Tests that the instantiation of the distribution raises an error."""
    dist_kwargs = dist_kwargs or {}
    try:
      with self.assertRaises(error_type):
        self.distrax_cls(*dist_args, **dist_kwargs)
    except ValueError:
      # For forward compatibility with Chex (it will raise AssertionErrors
      # instead of ValueErrors in the new version) .
      # TODO(iukemaev): remove after the new version of Chex is released.
      pass



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `block.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.bijectors import bijector as base
from distrax._src.bijectors import block as block_bijector
from distrax._src.bijectors import scalar_affine
from distrax._src.utils import conversion
import jax
import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp


tfd = tfp.distributions
tfb = tfp.bijectors


RTOL = 1e-6


class BlockTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.seed = jax.random.PRNGKey(1234)

  def test_properties(self):
    bijct = conversion.as_bijector(jnp.tanh)
    block = block_bijector.Block(bijct, 1)
    assert block.ndims == 1
    assert isinstance(block.bijector, base.Bijector)

  def test_invalid_properties(self):
    bijct = conversion.as_bijector(jnp.tanh)
    with self.assertRaises(ValueError):
      block_bijector.Block(bijct, -1)

  @chex.all_variants
  @parameterized.named_parameters(
      ('scale_0', lambda: tfb.Scale(2), 0),
      ('scale_1', lambda: tfb.Scale(2), 1),
      ('scale_2', lambda: tfb.Scale(2), 2),
      ('reshape_0', lambda: tfb.Reshape([120], [4, 5, 6]), 0),
      ('reshape_1', lambda: tfb.Reshape([120], [4, 5, 6]), 1),
      ('reshape_2', lambda: tfb.Reshape([120], [4, 5, 6]), 2),
  )
  def test_against_tfp_semantics(self, tfp_bijector_fn, ndims):
    tfp_bijector = tfp_bijector_fn()
    x = jax.random.normal(self.seed, [2, 3, 4, 5, 6])
    y = tfp_bijector(x)
    fwd_event_ndims = ndims + tfp_bijector.forward_min_event_ndims
    inv_event_ndims = ndims + tfp_bijector.inverse_min_event_ndims
    block = block_bijector.Block(tfp_bijector, ndims)
    np.testing.assert_allclose(
        tfp_bijector.forward_log_det_jacobian(x, fwd_event_ndims),
        self.variant(block.forward_log_det_jacobian)(x), atol=2e-5)
    np.testing.assert_allclose(
        tfp_bijector.inverse_log_det_jacobian(y, inv_event_ndims),
        self.variant(block.inverse_log_det_jacobian)(y), atol=2e-5)

  @chex.all_variants
  @parameterized.named_parameters(
      ('dx_tanh_0', lambda: jnp.tanh, 0),
      ('dx_tanh_1', lambda: jnp.tanh, 1),
      ('dx_tanh_2', lambda: jnp.tanh, 2),
      ('tfp_tanh_0', tfb.Tanh, 0),
      ('tfp_tanh_1', tfb.Tanh, 1),
      ('tfp_tanh_2', tfb.Tanh, 2),
  )
  def test_forward_inverse_work_as_expected(self, bijector_fn, ndims):
    bijct = conversion.as_bijector(bijector_fn())
    x = jax.random.normal(self.seed, [2, 3])
    block = block_bijector.Block(bijct, ndims)
    np.testing.assert_array_equal(
        self.variant(bijct.forward)(x),
        self.variant(block.forward)(x))
    np.testing.assert_array_equal(
        self.variant(bijct.inverse)(x),
        self.variant(block.inverse)(x))
    np.testing.assert_allclose(
        self.variant(bijct.forward_and_log_det)(x)[0],
        self.variant(block.forward_and_log_det)(x)[0], atol=2e-7)
    np.testing.assert_array_equal(
        self.variant(bijct.inverse_and_log_det)(x)[0],
        self.variant(block.inverse_and_log_det)(x)[0])

  @chex.all_variants
  @parameterized.named_parameters(
      ('dx_tanh_0', lambda: jnp.tanh, 0),
      ('dx_tanh_1', lambda: jnp.tanh, 1),
      ('dx_tanh_2', lambda: jnp.tanh, 2),
      ('tfp_tanh_0', tfb.Tanh, 0),
      ('tfp_tanh_1', tfb.Tanh, 1),
      ('tfp_tanh_2', tfb.Tanh, 2),
  )
  def test_log_det_jacobian_works_as_expected(self, bijector_fn, ndims):
    bijct = conversion.as_bijector(bijector_fn())
    x = jax.random.normal(self.seed, [2, 3])
    block = block_bijector.Block(bijct, ndims)
    axes = tuple(range(-ndims, 0))
    np.testing.assert_allclose(
        self.variant(bijct.forward_log_det_jacobian)(x).sum(axes),
        self.variant(block.forward_log_det_jacobian)(x), rtol=RTOL)
    np.testing.assert_allclose(
        self.variant(bijct.inverse_log_det_jacobian)(x).sum(axes),
        self.variant(block.inverse_log_det_jacobian)(x), rtol=RTOL)
    np.testing.assert_allclose(
        self.variant(bijct.forward_and_log_det)(x)[1].sum(axes),
        self.variant(block.forward_and_log_det)(x)[1], rtol=RTOL)
    np.testing.assert_allclose(
        self.variant(bijct.inverse_and_log_det)(x)[1].sum(axes),
        self.variant(block.inverse_and_log_det)(x)[1], rtol=RTOL)

  def test_raises_on_invalid_input_shape(self):
    bij = block_bijector.Block(lambda x: x, 1)
    for fn in [bij.forward, bij.inverse,
               bij.forward_log_det_jacobian, bij.inverse_log_det_jacobian,
               bij.forward_and_log_det, bij.inverse_and_log_det]:
      with self.assertRaises(ValueError):
        fn(jnp.array(0))

  def test_jittable(self):
    @jax.jit
    def f(x, b):
      return b.forward(x)

    bijector = block_bijector.Block(scalar_affine.ScalarAffine(0), 1)
    x = np.zeros((2, 3))
    f(x, bijector)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `quantized.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions import independent
from distrax._src.distributions import quantized
from distrax._src.distributions import uniform
from distrax._src.utils import equivalence
import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions


class QuantizedTFPUniform(equivalence.EquivalenceTest):
  """Class to test distrax quantized distribution against TFP.

  The quantized distribution takes a distribution as input. These can be either
  distrax or TFP distributions. In this test, we set the base distribution to
  be a TFP uniform one. Subsequent tests evaluate other combinations.
  """

  def _make_tfp_base_distribution(self):
    return tfd.Uniform(0., 100.)

  _make_distrax_base_distribution = _make_tfp_base_distribution

  def setUp(self):
    super().setUp()
    self._init_distr_cls(quantized.Quantized)
    self.tfd_base_distribution = self._make_tfp_base_distribution()
    self.distrax_base_distribution = self._make_distrax_base_distribution()

  def test_event_shape(self):
    super()._test_event_shape((self.distrax_base_distribution,),
                              dict(),
                              tfp_dist_args=(self.tfd_base_distribution,))

  def test_batch_shape(self):
    super()._test_batch_shape((self.distrax_base_distribution,),
                              dict(),
                              tfp_dist_args=(self.tfd_base_distribution,))

  def test_low_and_high(self):
    distr_params = (uniform.Uniform(0., 100.), 0.5, 90.5)
    dist = self.distrax_cls(*distr_params)
    self.assertion_fn(rtol=1e-2)(dist.low, 1.)
    self.assertion_fn(rtol=1e-2)(dist.high, 90.)

  @chex.all_variants
  @parameterized.named_parameters(
      ('empty shape', ()),
      ('int shape', 10),
      ('2-tuple shape', (10, 20)),
  )
  def test_sample_shape(self, sample_shape):
    super()._test_sample_shape((self.distrax_base_distribution,),
                               dict(),
                               tfp_dist_args=(self.tfd_base_distribution,),
                               sample_shape=sample_shape)

  @chex.all_variants
  def test_sample_dtype(self):
    dist = self.distrax_cls(self.distrax_base_distribution)
    samples = self.variant(dist.sample)(seed=self.key)
    self.assertEqual(dist.dtype, samples.dtype)
    self.assertEqual(dist.dtype, self.distrax_base_distribution.dtype)

  @chex.all_variants
  @parameterized.named_parameters(
      ('no cutoffs, no shape', (None, None), ()),
      ('noop cutoffs, no shape', (-10., 200.), ()),
      ('low cutoff, no shape', (10., None), ()),
      ('high cutoff, no shape', (None, 50.), ()),
      ('both cutoffs, no shape', (10., 50.), ()),
      ('both cutoffs, int shape', (10., 50.), 5),
      ('no cutoffs, 2-tuple shape', (None, None), (5, 4)),
      ('noop cutoffs, 2-tuple shape', (-10., 200.), (5, 4)),
      ('low cutoff, 2-tuple shape', (10., None), (5, 4)),
      ('high cutoff, 2-tuple shape', (None, 50.), (5, 4)),
      ('both cutoff, 2-tuple shape', (10., 50.), (5, 4)),
  )
  def test_sample_and_log_prob(self, distr_params, sample_shape):
    super()._test_sample_and_log_prob(
        dist_args=(self.distrax_base_distribution,) + distr_params,
        dist_kwargs=dict(),
        tfp_dist_args=(self.tfd_base_distribution,) + distr_params,
        sample_shape=sample_shape,
        assertion_fn=self.assertion_fn(rtol=1e-2))

  @chex.all_variants
  @parameterized.named_parameters(
      ('no cutoffs, integer', (None, None), 20),
      ('noop cutoffs, integer', (-10., 200.), 20),
      ('low cutoff, integer', (10., None), 20),
      ('high cutoff, integer', (None, 50.), 20),
      ('both cutoffs, integer', (10., 50.), 20),
      ('both cutoffs, integer greater than cutoff', (10., 50.), 70),
      ('both cutoffs, integer smaller than cutoff', (10., 50.), 5),
      ('both cutoffs, 1-d array', (10., 50.), np.array([20, 30])),
      ('no cutoffs, 1-d array', (None, None), np.array([20, 30])),
      ('noop cutoffs, 1-d array', (-10., 200.), np.array([20, 30])),
      ('low cutoffs, 1-d array', (10., None), np.array([20, 30])),
      ('high cutoffs, 1-d array', (None, 50.), np.array([20, 30])),
  )
  def test_method_with_value(self, distr_params, value):
    for method in ['log_cdf', 'cdf', 'prob', 'survival_function',
                   'log_survival_function']:
      with self.subTest(method):
        super()._test_attribute(
            attribute_string=method,
            dist_args=(self.distrax_base_distribution,) + distr_params,
            tfp_dist_args=(self.tfd_base_distribution,) + distr_params,
            call_args=(value,),
            assertion_fn=self.assertion_fn(rtol=1e-2))

  @chex.all_variants
  @parameterized.named_parameters(
      ('no cutoffs, integer', (None, None), 20),
      ('noop cutoffs, integer', (-10., 200.), 20),
      ('low cutoff, integer', (10., None), 20),
      ('high cutoff, integer', (None, 50.), 20),
      ('both cutoffs, integer', (10., 50.), 20),
      ('both cutoffs, 1-d array', (10., 50.), np.array([20, 30])),
      ('no cutoffs, 1-d array', (None, None), np.array([20, 30])),
      ('noop cutoffs, 1-d array', (-10., 200.), np.array([20, 30])),
      ('low cutoffs, 1-d array', (10., None), np.array([20, 30])),
      ('high cutoffs, 1-d array', (None, 50.), np.array([20, 30])),
  )
  def test_log_prob(self, distr_params, value):
    """Tests the `log_prob`.

    We separate this test from `test_method_with_value` because the options
    where `value` is outside the cutoff return `nan` in TFP but `-inf` in
    Distrax.

    Args:
      distr_params: Parameters of the distribution.
      value: The value where the `log_prob` is evaluated.
    """
    super()._test_attribute(
        attribute_string='log_prob',
        dist_args=(self.distrax_base_distribution,) + distr_params,
        tfp_dist_args=(self.tfd_base_distribution,) + distr_params,
        call_args=(value,),
        assertion_fn=self.assertion_fn(rtol=1e-2))

  @chex.all_variants
  @parameterized.named_parameters(
      ('below low', 10., 90., 8),
      ('above high', 10., 90., 95),
      ('below support', None, None, -10),
      ('above support', None, None, 101),
      ('within support, non-integer', None, None, 40.5),
  )
  def test_edge_cases(self, low, high, value):
    distr_params = {
        'distribution': self.distrax_base_distribution,
        'low': low,
        'high': high,
    }
    dist = self.distrax_cls(**distr_params)
    np.testing.assert_allclose(self.variant(dist.log_prob)(value), -np.inf)
    np.testing.assert_allclose(self.variant(dist.prob)(value), 0.)

  @parameterized.named_parameters(
      ('low with cutoffs', (10., 50.), 'low'),
      ('high with cutoffs', (10., 50.), 'high'),
  )
  def test_method(self, distr_params, function_string):
    super()._test_attribute(
        attribute_string=function_string,
        dist_args=(self.distrax_base_distribution,) + distr_params,
        dist_kwargs=dict(),
        tfp_dist_args=(self.tfd_base_distribution,) + distr_params,
        assertion_fn=self.assertion_fn(rtol=1e-2))

  def test_jittable(self):
    super()._test_jittable((self.tfd_base_distribution, 0., 1.))


class QuantizedDistraxUniform(QuantizedTFPUniform):

  def _make_distrax_base_distribution(self):
    return uniform.Uniform(0., 100.)

  def test_jittable(self):
    super()._test_jittable((self.distrax_base_distribution, 0., 1.))


class QuantizedTFPUniform2D(equivalence.EquivalenceTest):
  """Class to test distrax quantized distribution against TFP.

  The quantized distribution takes a distribution as input. These can be either
  distrax or TFP distributions. In this test, we set the base distribution to
  be a TFP uniform one with `batch_shape == (2,)`. Subsequent tests evaluate
  other combinations.
  """

  def _make_tfp_base_distribution(self):
    return tfd.Uniform(low=[0., 10.], high=[100., 90.])

  _make_distrax_base_distribution = _make_tfp_base_distribution

  def setUp(self):
    super().setUp()
    self._init_distr_cls(quantized.Quantized)
    self.tfd_base_distribution = self._make_tfp_base_distribution()
    self.distrax_base_distribution = self._make_distrax_base_distribution()

  def test_event_shape(self):
    kwargs = {
        'low': np.array([10., 30.], dtype=np.float32),
        'high': np.array([80., 70.], dtype=np.float32),
    }
    super()._test_event_shape(dist_args=(self.distrax_base_distribution,),
                              dist_kwargs=kwargs,
                              tfp_dist_args=(self.tfd_base_distribution,))

  def test_batch_shape(self):
    kwargs = {
        'low': np.array([10., 30.], dtype=np.float32),
        'high': np.array([80., 70.], dtype=np.float32),
    }
    super()._test_batch_shape(dist_args=(self.distrax_base_distribution,),
                              dist_kwargs=kwargs,
                              tfp_dist_args=(self.tfd_base_distribution,))

  @chex.all_variants
  @parameterized.named_parameters(
      ('empty shape', ()),
      ('int shape', 10),
      ('2-tuple shape', (10, 20)),
  )
  def test_sample_shape(self, sample_shape):
    kwargs = {
        'low': np.array([10., 30.], dtype=np.float32),
        'high': np.array([80., 70.], dtype=np.float32),
    }
    super()._test_sample_shape(dist_args=(self.distrax_base_distribution,),
                               dist_kwargs=kwargs,
                               tfp_dist_args=(self.tfd_base_distribution,),
                               sample_shape=sample_shape)

  @chex.all_variants
  @parameterized.named_parameters(
      ('no cutoffs, no shape', (None, None), ()),
      ('scalar cutoffs, no shape', (20., 50.), ()),
      ('1d cutoffs, no shape', ([10., 10.], [20., 80.]), ()),
      ('mixed cutoffs, no shape', ([10., 10.], 20.), ()),
      ('no cutoffs, int shape', (None, None), 10),
      ('scalar cutoffs, int shape', (20., 50.), 10),
      ('1d cutoffs, int shape', ([10., 10.], [20., 80.]), 10),
      ('mixed cutoffs, int shape', ([10., 10.], 20.), 10),
      ('no cutoffs, 1d shape', (None, None), [10]),
      ('scalar cutoffs, 1d shape', (20., 50.), [10]),
      ('1d cutoffs, 1d shape', ([10., 10.], [20., 80.]), [10]),
      ('mixed cutoffs, 1d shape', ([10., 10.], 20.), [10]),
  )
  def test_sample_and_log_prob(self, distr_params, sample_shape):
    distr_params = tuple(map(
        lambda x: None if x is None else np.asarray(x, np.float32),
        distr_params))
    super()._test_sample_and_log_prob(
        dist_args=(self.distrax_base_distribution,) + distr_params,
        dist_kwargs=dict(),
        tfp_dist_args=(self.tfd_base_distribution,) + distr_params,
        sample_shape=sample_shape,
        assertion_fn=self.assertion_fn(rtol=1e-2))

  @chex.all_variants
  @parameterized.named_parameters(
      ('no cutoffs, scalar value', (None, None), 20),
      ('scalar cutoffs, scalar value', (20., 50.), 20),
      ('1d cutoffs, scalar value', ([10., 12.], [20., 80.]), 15),
      ('mixed cutoffs, scalar value', ([10., 15.], 20.), 18),
      ('no cutoffs, 1d value', (None, None), np.array([20, 30])),
      ('scalar cutoffs, 1d value', (20., 50.), np.array([20, 30])),
      ('1d cutoffs, 1d value', ([10., 20.], [20., 80.]), np.array([20, 20])),
      ('mixed cutoffs, 1d value', ([10., 15.], 20.), np.array([11, 20])),
      ('mixed cutoffs, 2d value',
       ([10., 15.], 80.), np.array([[15, 15], [10, 20], [15, 15]])),
  )
  def test_method_with_value(self, distr_params, value):
    # For `prob`, `log_prob`, `survival_function`, and `log_survival_function`
    # distrax and TFP agree on integer values. We do not test equivalence on
    # non-integer values where they may disagree.
    # We also do not test equivalence on values outside of the cutoff, where
    # `log_prob` values can be different (`NaN` vs. `-jnp.inf`).
    distr_params = tuple(map(
        lambda x: None if x is None else np.asarray(x, np.float32),
        distr_params))
    for method in ['log_cdf', 'cdf', 'prob', 'log_prob', 'survival_function',
                   'log_survival_function']:
      with self.subTest(method):
        super()._test_attribute(
            attribute_string=method,
            dist_args=(self.distrax_base_distribution,) + distr_params,
            tfp_dist_args=(self.tfd_base_distribution,) + distr_params,
            call_args=(value,),
            assertion_fn=self.assertion_fn(rtol=1e-2))

  def test_jittable(self):
    super()._test_jittable((self.tfd_base_distribution, 0., 1.))


class QuantizedDistraxUniform2D(QuantizedTFPUniform2D):

  def _make_distrax_base_distribution(self):
    return uniform.Uniform(
        low=jnp.array([0.0, 10.0]), high=jnp.array([100.0, 90.0])
    )

  def test_jittable(self):
    super()._test_jittable((self.distrax_base_distribution, 0., 1.))


class QuantizedInvalidParams(equivalence.EquivalenceTest):
  """Class to test invalid combinations of the input parameters."""

  def setUp(self):
    super().setUp()
    self._init_distr_cls(quantized.Quantized)

  def test_non_univariate(self):
    self._test_raises_error(dist_kwargs={
        'distribution': independent.Independent(
            uniform.Uniform(np.array([0., 0.]), np.array([1., 1,])),
            reinterpreted_batch_ndims=1),
    })

  def test_low_shape(self):
    self._test_raises_error(dist_kwargs={
        'distribution': uniform.Uniform(0., 1.),
        'low': np.zeros((4,))
    })

  def test_high_shape(self):
    self._test_raises_error(dist_kwargs={
        'distribution': uniform.Uniform(0., 1.),
        'high': np.ones((4,))
    })


class QuantizedSlicingTest(parameterized.TestCase):
  """Class to test the `getitem` method."""

  def setUp(self):
    super().setUp()
    self.uniform_low = np.random.randn(2, 3, 4)
    self.uniform_high = self.uniform_low + np.abs(np.random.randn(2, 3, 4))
    self.base = uniform.Uniform(self.uniform_low, self.uniform_high)
    self.low = np.ceil(np.random.randn(3, 4))
    self.high = np.floor(np.random.randn(3, 4))
    self.dist = quantized.Quantized(self.base, self.low, self.high)

  def assertion_fn(self, rtol):
    return lambda x, y: np.testing.assert_allclose(x, y, rtol=rtol)

  @parameterized.named_parameters(
      ('single element', 1, (3, 4)),
      ('range', slice(-1), (1, 3, 4)),
      ('range_2', (slice(None), slice(-1)), (2, 2, 4)),
      ('ellipsis', (Ellipsis, -1), (2, 3)),
  )
  def test_slice(self, slice_, expected_batch_shape):
    sliced_dist = self.dist[slice_]
    self.assertEqual(sliced_dist.batch_shape, expected_batch_shape)
    self.assertEqual(sliced_dist.event_shape, self.dist.event_shape)
    self.assertIsInstance(sliced_dist, quantized.Quantized)
    self.assertIsInstance(sliced_dist.distribution, self.base.__class__)
    self.assertion_fn(rtol=1e-2)(
        sliced_dist.distribution.low, self.uniform_low[slice_])
    self.assertion_fn(rtol=1e-2)(
        sliced_dist.distribution.high, self.uniform_high[slice_])
    self.assertion_fn(rtol=1e-2)(
        sliced_dist.low,
        np.broadcast_to(self.low, self.base.batch_shape)[slice_])
    self.assertion_fn(rtol=1e-2)(
        sliced_dist.high,
        np.broadcast_to(self.high, self.base.batch_shape)[slice_])


class QuantizedSurvivalFunctionConsistencyTest(parameterized.TestCase):
  """Class to test whether `survival_function` = `1. - cdf`.

  Test evaluates on both integer values and non-integer values.
  """

  def setUp(self):
    super().setUp()
    self.base_distribution = uniform.Uniform(0., 10.)
    self.values = np.linspace(-2., 12, num=57)  # -2, -1.75, -1.5, ..., 12.

  @chex.all_variants
  @parameterized.named_parameters(
      ('no cutoffs', (None, None)),
      ('noop cutoffs', (-10., 20.)),
      ('low cutoff', (1., None)),
      ('high cutoff', (None, 5.)),
      ('both cutoffs', (1., 5.)),
  )
  def test_survival_function_cdf_consistency(self, dist_params):
    dist = quantized.Quantized(self.base_distribution, *dist_params)
    results = self.variant(
        lambda x: dist.cdf(x) + dist.survival_function(x))(self.values)
    np.testing.assert_allclose(results, np.ones_like(self.values), rtol=1e-2)

  @chex.all_variants
  @parameterized.named_parameters(
      ('no cutoffs', (None, None)),
      ('noop cutoffs', (-10., 20.)),
      ('low cutoff', (1., None)),
      ('high cutoff', (None, 5.)),
      ('both cutoffs', (1., 5.)),
  )
  def test_log_survival_function_log_cdf_consistency(self, dist_params):
    def _sum_exps(dist, x):
      return jnp.exp(dist.log_cdf(x)) + jnp.exp(dist.log_survival_function(x))
    dist = quantized.Quantized(self.base_distribution, *dist_params)
    results = self.variant(_sum_exps)(dist, self.values)
    np.testing.assert_allclose(results, np.ones_like(self.values), rtol=1e-2)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Abstract class for Jittable objects."""

import abc
import jax


class Jittable(metaclass=abc.ABCMeta):
  """ABC that can be passed as an arg to a jitted fn, with readable state."""

  def __new__(cls, *args, **kwargs):
    # Discard the parameters to this function because the constructor is not
    # called during serialization: its `__dict__` gets repopulated directly.
    del args, kwargs
    try:
      registered_cls = jax.tree_util.register_pytree_node_class(cls)
    except ValueError:
      registered_cls = cls  # Already registered.
    return object.__new__(registered_cls)

  def tree_flatten(self):
    leaves, treedef = jax.tree_util.tree_flatten(self.__dict__)
    switch = list(map(_is_jax_data, leaves))
    children = [leaf if s else None for leaf, s in zip(leaves, switch)]
    metadata = [None if s else leaf for leaf, s in zip(leaves, switch)]
    return children, (metadata, switch, treedef)

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    metadata, switch, treedef = aux_data
    leaves = [j if s else p for j, p, s in zip(children, metadata, switch)]
    obj = object.__new__(cls)
    obj.__dict__ = jax.tree_util.tree_unflatten(treedef, leaves)
    return obj


def _is_jax_data(x):
  """Check whether `x` is an instance of a JAX-compatible type."""
  # If it's a tracer, then it's already been converted by JAX.
  if isinstance(x, jax.core.Tracer):
    return True

  # `jax.vmap` replaces vmappable leaves with `object()` during serialization.
  if type(x) is object:  # pylint: disable=unidiomatic-typecheck
    return True

  # Primitive types (e.g. shape tuples) are treated as metadata for Distrax.
  if isinstance(x, (bool, int, float)) or x is None:
    return False

  # Return True if JAX considers `x` a valid JAX type.
  return jax.core.valid_jaxtype(x)



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Categorical distribution."""

from typing import Any, Optional, Tuple, Union

import chex
from distrax._src.distributions import distribution
from distrax._src.utils import math
import jax
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions

Array = chex.Array
PRNGKey = chex.PRNGKey
EventT = distribution.EventT


class Categorical(distribution.Distribution):
  """Categorical distribution."""

  equiv_tfp_cls = tfd.Categorical

  def __init__(self,
               logits: Optional[Array] = None,
               probs: Optional[Array] = None,
               dtype: Union[jnp.dtype, type[Any]] = int):
    """Initializes a Categorical distribution.

    Args:
      logits: Logit transform of the probability of each category. Only one
        of `logits` or `probs` can be specified.
      probs: Probability of each category. Only one of `logits` or `probs` can
        be specified.
      dtype: The type of event samples.
    """
    super().__init__()
    if (logits is None) == (probs is None):
      raise ValueError(
          f'One and exactly one of `logits` and `probs` should be `None`, '
          f'but `logits` is {logits} and `probs` is {probs}.')
    if not (jnp.issubdtype(dtype, jnp.integer) or
            jnp.issubdtype(dtype, jnp.floating)):
      raise ValueError(
          f'The dtype of `{self.name}` must be integer or floating-point, '
          f'instead got `{dtype}`.')

    self._probs = None if probs is None else math.normalize(probs=probs)
    self._logits = None if logits is None else math.normalize(logits=logits)
    self._dtype = dtype

  @property
  def event_shape(self) -> Tuple[int, ...]:
    """Shape of event of distribution samples."""
    return ()

  @property
  def logits(self) -> Array:
    """The logits for each event."""
    if self._logits is not None:
      return self._logits
    return jnp.log(self._probs)

  @property
  def probs(self) -> Array:
    """The probabilities for each event."""
    if self._probs is not None:
      return self._probs
    return jax.nn.softmax(self._logits, axis=-1)

  @property
  def num_categories(self) -> int:
    """Number of categories."""
    if self._probs is not None:
      return self._probs.shape[-1]
    return self._logits.shape[-1]

  def _sample_n(self, key: PRNGKey, n: int) -> Array:
    """See `Distribution._sample_n`."""
    new_shape = (n,) + self.logits.shape[:-1]
    is_valid = jnp.logical_and(jnp.all(jnp.isfinite(self.probs), axis=-1),
                               jnp.all(self.probs >= 0, axis=-1))
    draws = jax.random.categorical(key=key, logits=self.logits, axis=-1,
                                   shape=new_shape).astype(self._dtype)
    return jnp.where(is_valid, draws, jnp.ones_like(draws) * -1)

  def log_prob(self, value: EventT) -> Array:
    """See `Distribution.log_prob`."""
    value_one_hot = jax.nn.one_hot(
        value, self.num_categories, dtype=self.logits.dtype)
    mask_outside_domain = jnp.logical_or(
        value < 0, value > self.num_categories - 1)
    return jnp.where(
        mask_outside_domain, -jnp.inf,
        jnp.sum(math.multiply_no_nan(self.logits, value_one_hot), axis=-1))

  def prob(self, value: EventT) -> Array:
    """See `Distribution.prob`."""
    value_one_hot = jax.nn.one_hot(
        value, self.num_categories, dtype=self.probs.dtype)
    return jnp.sum(math.multiply_no_nan(self.probs, value_one_hot), axis=-1)

  def entropy(self) -> Array:
    """See `Distribution.entropy`."""
    if self._logits is None:
      log_probs = jnp.log(self._probs)
    else:
      log_probs = jax.nn.log_softmax(self._logits)
    return -jnp.sum(math.mul_exp(log_probs, log_probs), axis=-1)

  def mode(self) -> Array:
    """See `Distribution.mode`."""
    parameter = self._probs if self._logits is None else self._logits
    return jnp.argmax(parameter, axis=-1).astype(self._dtype)

  def cdf(self, value: EventT) -> Array:
    """See `Distribution.cdf`."""
    # For value < 0 the output should be zero because support = {0, ..., K-1}.
    should_be_zero = value < 0
    # For value >= K-1 the output should be one. Explicitly accounting for this
    # case addresses potential numerical issues that may arise when evaluating
    # derived methods (mainly, `log_survival_function`) for `value >= K-1`.
    should_be_one = value >= self.num_categories - 1
    # Will use value as an index below, so clip it to {0, ..., K-1}.
    value = jnp.clip(value, 0, self.num_categories - 1)
    value_one_hot = jax.nn.one_hot(
        value, self.num_categories, dtype=self.probs.dtype)
    cdf = jnp.sum(math.multiply_no_nan(
        jnp.cumsum(self.probs, axis=-1), value_one_hot), axis=-1)
    return jnp.where(should_be_zero, 0., jnp.where(should_be_one, 1., cdf))

  def log_cdf(self, value: EventT) -> Array:
    """See `Distribution.log_cdf`."""
    return jnp.log(self.cdf(value))

  def logits_parameter(self) -> Array:
    """Wrapper for `logits` property, for TFP API compatibility."""
    return self.logits

  def __getitem__(self, index) -> 'Categorical':
    """See `Distribution.__getitem__`."""
    index = distribution.to_batch_shape_index(self.batch_shape, index)
    if self._logits is not None:
      return Categorical(logits=self.logits[index], dtype=self._dtype)
    return Categorical(probs=self.probs[index], dtype=self._dtype)


CategoricalLike = Union[Categorical, tfd.Categorical]


def _kl_divergence_categorical_categorical(
    dist1: CategoricalLike,
    dist2: CategoricalLike,
    *unused_args, **unused_kwargs,
    ) -> Array:
  """Obtains the KL divergence `KL(dist1 || dist2)` between two Categoricals.

  The KL computation takes into account that `0 * log(0) = 0`; therefore,
  `dist1` may have zeros in its probability vector.

  Args:
    dist1: A Categorical distribution.
    dist2: A Categorical distribution.

  Returns:
    Batchwise `KL(dist1 || dist2)`.

  Raises:
    ValueError if the two distributions have different number of categories.
  """
  logits1 = dist1.logits_parameter()
  logits2 = dist2.logits_parameter()
  num_categories1 = logits1.shape[-1]
  num_categories2 = logits2.shape[-1]

  if num_categories1 != num_categories2:
    raise ValueError(
        f'Cannot obtain the KL between two Categorical distributions '
        f'with different number of categories: the first distribution has '
        f'{num_categories1} categories, while the second distribution has '
        f'{num_categories2} categories.')

  log_probs1 = jax.nn.log_softmax(logits1, axis=-1)
  log_probs2 = jax.nn.log_softmax(logits2, axis=-1)
  return jnp.sum(
      math.mul_exp(log_probs1 - log_probs2, log_probs1), axis=-1)


# Register the KL functions with TFP.
tfd.RegisterKL(Categorical, Categorical)(
    _kl_divergence_categorical_categorical)
tfd.RegisterKL(Categorical, Categorical.equiv_tfp_cls)(
    _kl_divergence_categorical_categorical)
tfd.RegisterKL(Categorical.equiv_tfp_cls, Categorical)(
    _kl_divergence_categorical_categorical)

# Also register the KL with the TFP OneHotCategorical.
tfd.RegisterKL(Categorical, tfd.OneHotCategorical)(
    _kl_divergence_categorical_categorical)
tfd.RegisterKL(tfd.OneHotCategorical, Categorical)(
    _kl_divergence_categorical_categorical)



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utility functions for performing Bijector-related JAX transformations.

A typical bijector will implement the following functionality:

- `forward`: the transformation of samples from an underlying distribution.
- `inverse`: the inverse function of `forward`.
- `[forward|inverse]_log_det_jacobian`: computes the log of the absolute value
    of the determinant of the Jacobian of the bijector's forward or inverse
    functions at a particular point.
- `is_constant_jacobian`: a boolean indicating whether the Jacobian of the
    forward and inverse functions is constant with respect to the function's
    input.

This module provides the following utilities for deriving these functions
automatically from JAX code implementing either the `forward` or `inverse`
functions:

- `inv`: inverts an input function `f` as long as it is composed of invertible
    primitives. This is achieved by tracing `f` to its underlying JAXPR
    representation and interpreting the code in reverse, substituting
    each invertible primitive by its inverse. Once transformed by `jit`, the
    code produced by the inversion interpreter is unrolled and JAX sees only the
    resulting inverse operations.
    The module provides the `register_inverse` function to specify custom
    inversion rules for primitives not yet supported here.
    See official JAX documentation for more information on tracing,
    writing custom interpreters, and a simple example of function inversion.
- `log_det_scalar`: computes the log-determinant of the Jacobian of a scalar
    function, using JAX autograd machinery.
- `is_constant_jacobian`: attempts to determine whether the Jacobian of `f` is
    constant with respect to its input by tracing `f` to its underlying JAXPR
    representation and checking whether any transformations of the input
    appear in the output. This is an experimental feature.
"""

import functools

from absl import logging
import jax
import jax.extend as jex
import jax.numpy as jnp

# pylint: disable=g-import-not-at-top
# pytype: disable=import-error
try:
  # jax >= 0.4.16
  from jax.extend import linear_util as lu
except ImportError:
  from jax import linear_util as lu
# pytype: enable=import-error
# pylint: enable=g-import-not-at-top


_inverse_registry = {
    # unary ops
    jax.lax.tanh_p: lambda x, **kwargs: jax.lax.atanh_p.bind(x),
    jax.lax.atanh_p: jax.lax.tanh_p,
    jax.lax.sinh_p: jax.lax.asinh_p,
    jax.lax.asinh_p: jax.lax.sinh_p,
    jax.lax.cosh_p: jax.lax.acosh_p,
    jax.lax.acosh_p: jax.lax.cosh_p,
    jax.lax.exp_p: jax.lax.log_p,
    jax.lax.log_p: jax.lax.exp_p,
    jax.lax.sqrt_p: lambda x, **kwargs: jax.lax.pow_p.bind(x, 2.0),
    jax.lax.rsqrt_p: lambda x, **kwargs: 1.0 / jax.lax.pow_p.bind(x, 2.0),
    jax.lax.neg_p: jax.lax.neg_p,
    jax.lax.log1p_p: jax.lax.expm1_p,
    jax.lax.expm1_p: jax.lax.log1p_p,
    jax.lax.erf_p: jax.lax.erf_inv_p,
    jax.lax.erf_inv_p: jax.lax.erf_p,
    jax.lax.conj_p: jax.lax.conj_p,

    # binary ops; tuple values represent the variable-left/variable-right side
    # case for non-commutatively invertible ops like div
    jax.lax.mul_p: (jax.lax.div_p.bind, lambda x, y: jax.lax.div_p.bind(y, x)),
    jax.lax.div_p: (jax.lax.mul_p.bind, jax.lax.div_p.bind),
    jax.lax.add_p: (jax.lax.sub_p.bind, lambda x, y: jax.lax.sub_p.bind(y, x)),
    jax.lax.sub_p: (jax.lax.add_p.bind, jax.lax.sub_p.bind),
    jax.lax.pow_p: lambda x, y: jax.lax.pow_p.bind(x, 1.0/y),
    jax.lax.integer_pow_p: lambda x, y: jax.lax.pow_p.bind(x, 1.0/y)
}

if hasattr(jax.lax, "square_p"):
  _inverse_registry.update({
      jax.lax.square_p: jax.lax.sqrt_p,
      jax.lax.sqrt_p: lambda x, **kwargs: jax.lax.square_p.bind(x),
      jax.lax.rsqrt_p: lambda x, **kwargs: 1.0 / jax.lax.square_p.bind(x),
  })

_potentially_unstable_primitives = {
    jax.lax.tanh_p: "distrax.Tanh or distrax.Inverse(distrax.Tanh)",
    jax.lax.atanh_p: "distrax.Tanh or distrax.Inverse(distrax.Tanh)",
}


def register_inverse(primitive, inverse_left, inverse_right=None):
  """Register a function that implements the inverse of a JAX primitive.

  Args:
    primitive: JAX primitive, often named `*_p` and located in `jax.lax.lax.py`.
    inverse_left: a function implementing the inverse if the primitive is
        a unary operator or if `inv(f(x,y)) == inv(f(y,x))`, else a function
        implementing the inverse of a binary operator when the variable in
        question comes before the operator, e.g. `x div_p 2`.
    inverse_right: a function implementing the inverse of a binary
        operator when the variable in question comes after the operator,
        e.g. `2 div_p x`.
  """
  if inverse_right is None:
    _inverse_registry[primitive] = inverse_left
  else:
    _inverse_registry[primitive] = (inverse_left, inverse_right)


def inv(fun):
  """Returns the inverse of `fun` such that (inv(fun) o fun)(x) = x."""
  jaxpr_fn = _invertible_jaxpr_and_constants(fun)

  @functools.wraps(fun)  # pylint: disable=no-value-for-parameter
  def wrapped(*args, **kwargs):
    jaxpr, consts = jaxpr_fn(*args, **kwargs)
    out = _interpret_inverse(jaxpr, consts, *args)
    return out[0]
  return wrapped


def is_constant_jacobian(fn, x=0.0):
  """Experimental. Attempts to determine whether `fn` has a constant Jacobian.

  This function attempts to determine whether the Jacobian of `fn` is constant
  w.r.t. its input. We compute the Jacobian of `fn` at `x` and inspect the
  jaxpr to determine whether any function of the input appears at the output.

  Args:
    fn: a JAX-traceable differentiable function taking scalar input.
    x: the location at which to check whether the Jacobian is constant.

  Returns:
    Boolean value indicating whether the Jacobian is constant at `x`.
  """

  jac_fn = jax.jacfwd(fn)
  jac_jaxpr = jax.make_jaxpr(jac_fn)(jnp.array(x)).jaxpr
  dependent_vars = _dependent_variables(jac_jaxpr)

  jac_is_constant = not any(isinstance(v, jex.core.Var) and v in dependent_vars
                            for v in jac_jaxpr.outvars)

  return jac_is_constant


def log_det_scalar(fn):
  """Uses JAX autograd to derive the log-det-jacobian of a scalar function."""
  _check_numerical_stability(fn)
  jac_fn = jax.vmap(jax.jacfwd(fn))
  def log_det_fn(x):
    x = jnp.asarray(x)
    jac_scalar = jac_fn(x.reshape(-1))
    log_det_ = jnp.log(jnp.absolute(jac_scalar))
    return log_det_.reshape(x.shape)
  return log_det_fn


def _check_numerical_stability(fn):
  """Logs a warning if numerically unstable operations are requested."""
  jaxpr = jax.make_jaxpr(fn)(0.0).jaxpr
  for eqn in jaxpr.eqns:
    if eqn.primitive in _potentially_unstable_primitives:
      logging.warn("[Distrax]: the '%s' primitive can exhibit unstable "
                   "numerical behavior under certain circumstances. Consider "
                   "using the %s bijector instead if possible.", eqn.primitive,
                   _potentially_unstable_primitives[eqn.primitive])


def _dependent_variables(jaxpr, dependent=None):
  """Returns the set of variables in the jaxpr that depend on the input vars."""
  if dependent is None:
    dependent = set(jaxpr.invars)

  for eqn in jaxpr.eqns:
    # If primitive is an xla_call, get subexpressions and evaluate recursively
    call_jaxpr, _ = _extract_call_jaxpr(eqn.primitive, eqn.params)
    if call_jaxpr:
      to_name = dict(zip(eqn.invars, call_jaxpr.invars))
      arg_dependence = set(to_name[v] for v in eqn.invars if v in dependent)
      subjaxpr_dependent = _dependent_variables(call_jaxpr, arg_dependence)
      from_name = dict(zip(call_jaxpr.outvars, eqn.outvars))
      dependent.update(from_name[v] for v in call_jaxpr.outvars
                       if v in subjaxpr_dependent)
    else:
      for v in eqn.invars:
        if isinstance(v, jex.core.Var) and v in dependent:
          dependent.update(eqn.outvars)

  return dependent


def _invertible_jaxpr_and_constants(fun):
  """Returns a transformation from function invocation to invertible jaxpr."""
  jaxpr_maker = jax.make_jaxpr(fun)

  @functools.wraps(fun)  # pylint: disable=no-value-for-parameter
  def jaxpr_const_maker(*args, **kwargs):
    typed_jaxpr = jaxpr_maker(*args, **kwargs)
    return typed_jaxpr.jaxpr, typed_jaxpr.literals
  return jaxpr_const_maker


def _identify_variable_in_eqn(eqn):
  """Identify whether primitive is a unop or binop and which side var is on."""

  if len(eqn.invars) == 1:  # unary operation
    var_idx = 0

  elif len(eqn.invars) == 2:  # binary operation
    if tuple(map(type, eqn.invars)) == (jex.core.Var, jex.core.Literal):
      var_idx = 0

    elif tuple(map(type, eqn.invars)) == (jex.core.Literal, jex.core.Var):
      var_idx = 1

    elif tuple(map(type, eqn.invars)) == (jex.core.Var, jex.core.Var):
      raise NotImplementedError(
          "Expressions with multiple occurrences of the input variable are "
          "not supported. Please rearrange such that the variable appears only "
          "once in the expression if possible. If not possible, consider "
          "providing both `forward` and `inverse` to Lambda explicitly.")

    elif tuple(map(type, eqn.invars)) == (jex.core.Literal, jex.core.Literal):
      raise ValueError("Expression appears to contain no variables and "
                       "therefore cannot be inverted.")

    else:
      raise NotImplementedError("Unsupported binary op combination: "
                                + str(tuple(map(type, eqn.invars))))

  else:
    raise NotImplementedError(f"Op {eqn.primitive} with cardinality >= 3 not "
                              "supported.")

  return var_idx


def _interpret_inverse(jaxpr, consts, *args):
  """Interprets and executes the inverse of `jaxpr`."""
  env = {}

  def read(var):
    return var.val if isinstance(var, jex.core.Literal) else env[var]
  def write(var, val):
    env[var] = val

  jax.api_util.safe_map(write, jaxpr.outvars, args)
  jax.api_util.safe_map(write, jaxpr.constvars, consts)

  for eqn in reversed(jaxpr.eqns):
    params = eqn.params.copy()
    # identify the cardinality of the op and the index of the variable in eqn
    var_idx = _identify_variable_in_eqn(eqn)

    # if primitive is an xla_call, get subexpressions and evaluate recursively
    call_jaxpr, params = _extract_call_jaxpr(eqn.primitive, params)
    if call_jaxpr:
      subfuns = [lu.wrap_init(
          functools.partial(_interpret_inverse, call_jaxpr, ()))]
      prim_inv = eqn.primitive

    elif eqn.primitive is jex.core.primitives.jit_p:
      pjit_jaxpr = params.pop("jaxpr")
      partial_inverse = functools.partial(_interpret_inverse, pjit_jaxpr.jaxpr,
                                          pjit_jaxpr.consts)
      inverse_jaxpr = jax.make_jaxpr(partial_inverse)(*args)
      params["jaxpr"] = inverse_jaxpr

      prim_inv = eqn.primitive
      subfuns = []

    else:  # otherwise, get its inverse if it exists
      if eqn.primitive not in _inverse_registry:
        raise NotImplementedError(
            f"Primitive '{eqn.primitive}' does not have a registered inverse.")

      # use the correct inverse formulation depending on whether the variable is
      # on the left or right side of the expression
      prim_inv = _inverse_registry[eqn.primitive]
      if isinstance(prim_inv, tuple):
        prim_inv = prim_inv[var_idx]

      subfuns = []

    # get the values of any variables in the eqn
    invals = jax.api_util.safe_map(read, eqn.outvars)

    # place the args and variables in the right order
    if var_idx == 0:
      prim_args = subfuns + invals + [v.val for v in eqn.invars[1:]]
    else:
      prim_args = subfuns + [v.val for v in eqn.invars[:1]] + invals

    # if the inverse is a primitive, bind it, otherwise call it directly
    if hasattr(prim_inv, "bind"):
      outvals = prim_inv.bind(*prim_args, **params)
    else:
      outvals = prim_inv(*prim_args, **params)

    # if the primitive returns multiple results, write them all to env
    if (hasattr(prim_inv, "multiple_results") and prim_inv.multiple_results):
      jax.api_util.safe_map(write, eqn.invars, outvals)
    else:
      write(eqn.invars[var_idx], outvals)

  if any(v not in env for v in jaxpr.invars):
    raise ValueError("Expression appears to contain no variables and therefore "
                     "cannot be inverted.")

  return jax.api_util.safe_map(read, jaxpr.invars)


def _extract_call_jaxpr(primitive, params):
  if not (primitive.call_primitive or primitive.map_primitive):
    return None, params
  else:
    params = dict(params)
    return params.pop("call_jaxpr"), params



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Diagonal linear bijector."""

from typing import Tuple

from distrax._src.bijectors import bijector as base
from distrax._src.bijectors import block
from distrax._src.bijectors import linear
from distrax._src.bijectors import scalar_affine
import jax.numpy as jnp

Array = base.Array


class DiagLinear(linear.Linear):
  """Linear bijector with a diagonal weight matrix.

  The bijector is defined as `f(x) = Ax` where `A` is a `DxD` diagonal matrix.
  Additional dimensions, if any, index batches.

  The Jacobian determinant is trivially computed by taking the product of the
  diagonal entries in `A`. The inverse transformation `x = f^{-1}(y)` is
  computed element-wise.

  The bijector is invertible if and only if the diagonal entries of `A` are all
  non-zero. It is the responsibility of the user to make sure that this is the
  case; the class will make no attempt to verify that the bijector is
  invertible.
  """

  def __init__(self, diag: Array):
    """Initializes the bijector.

    Args:
      diag: a vector of length D, the diagonal of matrix `A`. Can also be a
        batch of such vectors.
    """
    if diag.ndim < 1:
      raise ValueError("`diag` must have at least one dimension.")
    self._bijector = block.Block(
        scalar_affine.ScalarAffine(shift=0., scale=diag), ndims=1)
    super().__init__(
        event_dims=diag.shape[-1],
        batch_shape=diag.shape[:-1],
        dtype=diag.dtype)
    self._diag = diag
    self.forward = self._bijector.forward
    self.forward_log_det_jacobian = self._bijector.forward_log_det_jacobian
    self.inverse = self._bijector.inverse
    self.inverse_log_det_jacobian = self._bijector.inverse_log_det_jacobian
    self.inverse_and_log_det = self._bijector.inverse_and_log_det

  @property
  def diag(self) -> Array:
    """Vector of length D, the diagonal of matrix `A`."""
    return self._diag

  @property
  def matrix(self) -> Array:
    """The full matrix `A`."""
    return jnp.vectorize(jnp.diag, signature="(k)->(k,k)")(self.diag)

  def forward_and_log_det(self, x: Array) -> Tuple[Array, Array]:
    """Computes y = f(x) and log|det J(f)(x)|."""
    return self._bijector.forward_and_log_det(x)

  def same_as(self, other: base.Bijector) -> bool:
    """Returns True if this bijector is guaranteed to be the same as `other`."""
    if type(other) is DiagLinear:  # pylint: disable=unidiomatic-typecheck
      return self.diag is other.diag
    return False



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Mixture distributions."""

from typing import Tuple

import chex
from distrax._src.distributions import categorical
from distrax._src.distributions import distribution
from distrax._src.utils import conversion
import jax
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp


tfd = tfp.distributions

Array = chex.Array
PRNGKey = chex.PRNGKey
DistributionLike = distribution.DistributionLike
CategoricalLike = categorical.CategoricalLike
EventT = distribution.EventT


class MixtureSameFamily(distribution.Distribution):
  """Mixture with components provided from a single batched distribution."""

  equiv_tfp_cls = tfd.MixtureSameFamily

  def __init__(self,
               mixture_distribution: CategoricalLike,
               components_distribution: DistributionLike):
    """Initializes a mixture distribution for components of a shared family.

    Args:
      mixture_distribution: Distribution over selecting components.
      components_distribution: Component distribution, with rightmost batch
        dimension indexing components.
    """
    super().__init__()
    mixture_distribution = conversion.as_distribution(mixture_distribution)
    components_distribution = conversion.as_distribution(
        components_distribution)
    self._mixture_distribution = mixture_distribution
    self._components_distribution = components_distribution

    # Store normalized weights (last axis of logits is for components).
    # This uses the TFP API, which is replicated in Distrax.
    self._mixture_log_probs = jax.nn.log_softmax(
        mixture_distribution.logits_parameter(), axis=-1)

    batch_shape_mixture = mixture_distribution.batch_shape
    batch_shape_components = components_distribution.batch_shape
    if batch_shape_mixture != batch_shape_components[:-1]:
      msg = (f'`mixture_distribution.batch_shape` '
             f'({mixture_distribution.batch_shape}) is not compatible with '
             f'`components_distribution.batch_shape` '
             f'({components_distribution.batch_shape}`)')
      raise ValueError(msg)

  @property
  def components_distribution(self):
    """The components distribution."""
    return self._components_distribution

  @property
  def mixture_distribution(self):
    """The mixture distribution."""
    return self._mixture_distribution

  @property
  def event_shape(self) -> Tuple[int, ...]:
    """Shape of event of distribution samples."""
    return self._components_distribution.event_shape

  @property
  def batch_shape(self) -> Tuple[int, ...]:
    """Shape of batch of distribution samples."""
    return self._components_distribution.batch_shape[:-1]

  def _sample_n(self, key: PRNGKey, n: int) -> Array:
    """See `Distribution._sample_n`."""
    key_mix, key_components = jax.random.split(key)
    mix_sample = self.mixture_distribution.sample(sample_shape=n, seed=key_mix)

    num_components = self._components_distribution.batch_shape[-1]

    # Sample from all components, then multiply with a one-hot mask and sum.
    # While this does computation that is not used eventually, it is faster on
    # GPU/TPUs, which excel at batched operations (as opposed to indexing). It
    # is in particular more efficient than using `gather` or `where` operations.
    mask = jax.nn.one_hot(mix_sample, num_components)
    samples_all = self.components_distribution.sample(sample_shape=n,
                                                      seed=key_components)

    # Make mask broadcast with (potentially multivariate) samples.
    mask = mask.reshape(mask.shape + (1,) * len(self.event_shape))

    # Need to sum over the component axis, which is the last one for scalar
    # components, the second-last one for 1-dim events, etc.
    samples = jnp.sum(samples_all * mask, axis=-1 - len(self.event_shape))
    return samples

  def log_prob(self, value: EventT) -> Array:
    """See `Distribution.log_prob`."""
    # Add component axis to make input broadcast with components distribution.
    expanded = jnp.expand_dims(value, axis=-1 - len(self.event_shape))
    # Compute `log_prob` in every component.
    lp = self.components_distribution.log_prob(expanded)
    # Last batch axis is number of components, i.e. last axis of `lp` below.
    # Last axis of mixture log probs are components, so reduce last axis.
    return jax.scipy.special.logsumexp(a=lp + self._mixture_log_probs, axis=-1)

  def mean(self) -> Array:
    """Calculates the mean."""
    means = self.components_distribution.mean()
    weights = jnp.exp(self._mixture_log_probs)
    # Broadcast weights over event shape, and average over component axis.
    weights = weights.reshape(weights.shape + (1,) * len(self.event_shape))
    return jnp.sum(means * weights, axis=-1 - len(self.event_shape))

  def variance(self) -> Array:
    """Calculates the variance."""
    means = self.components_distribution.mean()
    variances = self.components_distribution.variance()
    weights = jnp.exp(self._mixture_log_probs)
    # Make weights broadcast over event shape.
    weights = weights.reshape(weights.shape + (1,) * len(self.event_shape))
    # Component axis to reduce over.
    component_axis = -1 - len(self.event_shape)

    # Using: Var(Y) = E[Var(Y|X)] + Var(E[Y|X]).
    mean = jnp.sum(means * weights, axis=component_axis)
    mean_cond_var = jnp.sum(weights * variances, axis=component_axis)
    # Need to add an axis to `mean` to make it broadcast over components.
    sq_diff = jnp.square(means - jnp.expand_dims(mean, axis=component_axis))
    var_cond_mean = jnp.sum(weights * sq_diff, axis=component_axis)
    return mean_cond_var + var_cond_mean

  def __getitem__(self, index) -> 'MixtureSameFamily':
    """See `Distribution.__getitem__`."""
    index = distribution.to_batch_shape_index(self.batch_shape, index)
    return MixtureSameFamily(
        mixture_distribution=self.mixture_distribution[index],
        components_distribution=self.components_distribution[index])



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Deterministic distribution."""

from typing import Optional, Tuple, Union

import chex
from distrax._src.distributions import distribution
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp


tfd = tfp.distributions

Array = chex.Array
Numeric = chex.Numeric
PRNGKey = chex.PRNGKey
EventT = distribution.EventT


class Deterministic(distribution.Distribution):
  """Scalar Deterministic distribution on the real line."""

  equiv_tfp_cls = tfd.Deterministic

  def __init__(self,
               loc: Numeric,
               atol: Optional[Numeric] = None,
               rtol: Optional[Numeric] = None):
    """Initializes a Deterministic distribution.

    Args:
      loc: Batch of points on which the distribution is supported.
      atol: Absolute tolerance for comparing closeness to `loc`. It must be
        broadcastable with `loc`, and it must not lead to additional batch
        dimensions after broadcasting.
      rtol: Relative tolerance for comparing closeness to `loc`. It must be
        broadcastable with `loc`, and it must not lead to additional batch
        dimensions after broadcasting.
    """
    super().__init__()
    self._loc = jnp.asarray(loc)
    self._atol = jnp.asarray(0. if atol is None else atol)
    self._rtol = jnp.asarray(0. if rtol is None else rtol)
    if len(self._rtol.shape) > len(self._loc.shape):
      raise ValueError(f'The parameter `rtol` cannot have more dimensions than '
                       f'`loc`, but their shapes are {self._rtol.shape} and '
                       f'{self._loc.shape}, respectively.')
    if len(self._atol.shape) > len(self._loc.shape):
      raise ValueError(f'The parameter `atol` cannot have more dimensions than '
                       f'`loc`, but their shapes are {self._atol.shape} and '
                       f'{self._loc.shape}, respectively.')

  @property
  def event_shape(self) -> Tuple[int, ...]:
    """Shape of the events."""
    return ()

  @property
  def batch_shape(self) -> Tuple[int, ...]:
    """Shape of batch of distribution samples."""
    return self._loc.shape

  @property
  def loc(self) -> Array:
    """Point(s) on which this distribution is supported."""
    return self._loc

  @property
  def atol(self) -> Array:
    """Absolute tolerance for comparing closeness to `loc`."""
    return jnp.broadcast_to(self._atol, self.batch_shape)

  @property
  def rtol(self) -> Array:
    """Relative tolerance for comparing closeness to `loc`."""
    return jnp.broadcast_to(self._rtol, self.batch_shape)

  @property
  def slack(self) -> Array:
    return jnp.where(
        self.rtol == 0,
        self.atol,
        self.atol + self.rtol * jnp.abs(self.loc))

  def _sample_n(self, key: PRNGKey, n: int) -> Array:
    """See `Distribution._sample_n`."""
    del key  # unused
    loc = jnp.expand_dims(self.loc, axis=0)
    return jnp.repeat(loc, n, axis=0)

  def _sample_n_and_log_prob(self, key: PRNGKey, n: int) -> Tuple[Array, Array]:
    """See `Distribution._sample_n_and_log_prob`."""
    samples = self._sample_n(key, n)
    log_prob = jnp.zeros_like(samples)
    return samples, log_prob

  def log_prob(self, value: EventT) -> Array:
    """See `Distribution.log_prob`."""
    return jnp.log(self.prob(value))

  def prob(self, value: EventT) -> Array:
    """See `Distribution.prob`."""
    return jnp.where(
        jnp.abs(value - self.loc) <= self.slack, 1., 0.)

  def entropy(self) -> Array:
    """Calculates the Shannon entropy (in nats)."""
    return jnp.zeros(self.batch_shape, jnp.float_)

  def mean(self) -> Array:
    """Calculates the mean."""
    return self.loc

  def mode(self) -> Array:
    """Calculates the mode."""
    return self.mean()

  def variance(self) -> Array:
    """Calculates the variance."""
    return jnp.zeros(self.batch_shape, jnp.float_)

  def stddev(self) -> Array:
    """Calculates the standard deviation."""
    return self.variance()

  def log_cdf(self, value: EventT) -> Array:
    """See `Distribution.log_cdf`."""
    return jnp.log(self.cdf(value))

  def cdf(self, value: EventT) -> Array:
    """See `Distribution.cdf`."""
    return jnp.where(value >= self.loc - self.slack, 1., 0.)

  def __getitem__(self, index) -> 'Deterministic':
    """See `Distribution.__getitem__`."""
    index = distribution.to_batch_shape_index(self.batch_shape, index)
    return Deterministic(
        loc=self.loc[index], atol=self.atol[index], rtol=self.rtol[index])


def _kl_divergence_deterministic_deterministic(
    dist1: Union[Deterministic, tfd.Deterministic],
    dist2: Union[Deterministic, tfd.Deterministic],
    *unused_args, **unused_kwargs,
    ) -> Array:
  """KL divergence `KL(dist1 || dist2)` between two Deterministic distributions.

  Note that the KL divergence is infinite if the support of `dist1` is not a
  subset of the support of `dist2`.

  Args:
    dist1: A Deterministic distribution.
    dist2: A Deterministic distribution.

  Returns:
    Batchwise `KL(dist1 || dist2)`.
  """
  slack2 = dist2.atol + dist2.rtol * jnp.abs(dist2.loc)
  return - jnp.log(jnp.where(jnp.abs(dist1.loc - dist2.loc) <= slack2, 1., 0.))


# Register the KL functions with TFP.
tfd.RegisterKL(Deterministic, Deterministic)(
    _kl_divergence_deterministic_deterministic)
tfd.RegisterKL(Deterministic, Deterministic.equiv_tfp_cls)(
    _kl_divergence_deterministic_deterministic)
tfd.RegisterKL(Deterministic.equiv_tfp_cls, Deterministic)(
    _kl_divergence_deterministic_deterministic)



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `jittable.py`."""

from absl.testing import absltest
from absl.testing import parameterized

from distrax._src.utils import jittable
import jax
import jax.numpy as jnp
import numpy as np


class DummyJittable(jittable.Jittable):

  def __init__(self, params):
    self.name = 'dummy'  # Non-JAX property, cannot be traced.
    self.data = {'params': params}  # Tree property, must be traced recursively.


class JittableTest(parameterized.TestCase):

  def test_jittable(self):
    @jax.jit
    def get_params(obj):
      return obj.data['params']
    obj = DummyJittable(jnp.ones((5,)))
    np.testing.assert_array_equal(get_params(obj), obj.data['params'])

  def test_vmappable(self):
    def do_sum(obj):
      return obj.data['params'].sum()
    obj = DummyJittable(jnp.array([[1, 2, 3], [4, 5, 6]]))

    with self.subTest('no vmap'):
      np.testing.assert_array_equal(do_sum(obj), obj.data['params'].sum())

    with self.subTest('in_axes=0'):
      np.testing.assert_array_equal(
          jax.vmap(do_sum, in_axes=0)(obj), obj.data['params'].sum(axis=1))

    with self.subTest('in_axes=1'):
      np.testing.assert_array_equal(
          jax.vmap(do_sum, in_axes=1)(obj), obj.data['params'].sum(axis=0))

  def test_traceable(self):
    @jax.jit
    def inner_fn(obj):
      obj.data['params'] *= 3  # Modification after passing to jitted fn.
      return obj.data['params'].sum()

    def loss_fn(params):
      obj = DummyJittable(params)
      obj.data['params'] *= 2  # Modification before passing to jitted fn.
      return inner_fn(obj)

    with self.subTest('numpy'):
      params = np.ones((5,))
      # Both modifications will be traced if data tree is correctly traversed.
      grad_expected = params * 2 * 3
      grad = jax.grad(loss_fn)(params)
      np.testing.assert_array_equal(grad, grad_expected)

    with self.subTest('jax.numpy'):
      params = jnp.ones((5,))
      # Both modifications will be traced if data tree is correctly traversed.
      grad_expected = params * 2 * 3
      grad = jax.grad(loss_fn)(params)
      np.testing.assert_array_equal(grad, grad_expected)

  def test_different_jittables_to_compiled_function(self):
    @jax.jit
    def add_one_to_params(obj):
      obj.data['params'] = obj.data['params'] + 1
      return obj

    with self.subTest('numpy'):
      add_one_to_params(DummyJittable(np.zeros((5,))))
      add_one_to_params(DummyJittable(np.ones((5,))))

    with self.subTest('jax.numpy'):
      add_one_to_params(DummyJittable(jnp.zeros((5,))))
      add_one_to_params(DummyJittable(jnp.ones((5,))))

  def test_modifying_object_data_does_not_leak_tracers(self):
    @jax.jit
    def add_one_to_params(obj):
      obj.data['params'] = obj.data['params'] + 1
      return obj

    dummy = DummyJittable(jnp.ones((5,)))
    dummy_out = add_one_to_params(dummy)
    dummy_out.data['params'] -= 1

  def test_metadata_modification_statements_are_removed_by_compilation(self):
    @jax.jit
    def add_char_to_name(obj):
      obj.name += '_x'
      return obj

    dummy = DummyJittable(jnp.ones((5,)))
    dummy_out = add_char_to_name(dummy)
    dummy_out = add_char_to_name(dummy)  # `name` change has been compiled out.
    dummy_out.name += 'y'
    self.assertEqual(dummy_out.name, 'dummy_xy')


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Logistic distribution."""

from typing import Tuple

import chex
from distrax._src.distributions import distribution
from distrax._src.utils import conversion
import jax
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions

Array = chex.Array
Numeric = chex.Numeric
PRNGKey = chex.PRNGKey
EventT = distribution.EventT


class Logistic(distribution.Distribution):
  """The Logistic distribution with location `loc` and `scale` parameters."""

  equiv_tfp_cls = tfd.Logistic

  def __init__(self, loc: Numeric, scale: Numeric) -> None:
    """Initializes a Logistic distribution.

    Args:
      loc: Mean of the distribution.
      scale: Spread of the distribution.
    """
    super().__init__()
    self._loc = conversion.as_float_array(loc)
    self._scale = conversion.as_float_array(scale)
    self._batch_shape = jax.lax.broadcast_shapes(
        self._loc.shape, self._scale.shape)

  @property
  def event_shape(self) -> Tuple[int, ...]:
    """Shape of event of distribution samples."""
    return ()

  @property
  def batch_shape(self) -> Tuple[int, ...]:
    """Shape of batch of distribution samples."""
    return self._batch_shape

  @property
  def loc(self) -> Array:
    """Mean of the distribution."""
    return jnp.broadcast_to(self._loc, self.batch_shape)

  @property
  def scale(self) -> Array:
    """Spread of the distribution."""
    return jnp.broadcast_to(self._scale, self.batch_shape)

  def _standardize(self, x: Array) -> Array:
    return (x - self.loc) / self.scale

  def _sample_n(self, key: PRNGKey, n: int) -> Array:
    """See `Distribution._sample_n`."""
    out_shape = (n,) + self.batch_shape
    dtype = jnp.result_type(self._loc, self._scale)
    uniform = jax.random.uniform(
        key,
        shape=out_shape,
        dtype=dtype,
        minval=jnp.finfo(dtype).tiny,
        maxval=1.)
    rnd = jnp.log(uniform) - jnp.log1p(-uniform)
    return self._scale * rnd + self._loc

  def log_prob(self, value: EventT) -> Array:
    """See `Distribution.log_prob`."""
    z = self._standardize(value)
    return -z - 2. * jax.nn.softplus(-z) - jnp.log(self._scale)

  def entropy(self) -> Array:
    """Calculates the Shannon entropy (in Nats)."""
    return 2. + jnp.broadcast_to(jnp.log(self._scale), self.batch_shape)

  def cdf(self, value: EventT) -> Array:
    """See `Distribution.cdf`."""
    return jax.nn.sigmoid(self._standardize(value))

  def log_cdf(self, value: EventT) -> Array:
    """See `Distribution.log_cdf`."""
    return -jax.nn.softplus(-self._standardize(value))

  def survival_function(self, value: EventT) -> Array:
    """See `Distribution.survival_function`."""
    return jax.nn.sigmoid(-self._standardize(value))

  def log_survival_function(self, value: EventT) -> Array:
    """See `Distribution.log_survival_function`."""
    return -jax.nn.softplus(self._standardize(value))

  def mean(self) -> Array:
    """Calculates the mean."""
    return self.loc

  def variance(self) -> Array:
    """Calculates the variance."""
    return jnp.square(self.scale * jnp.pi) / 3.

  def stddev(self) -> Array:
    """Calculates the standard deviation."""
    return self.scale * jnp.pi / jnp.sqrt(3.)

  def mode(self) -> Array:
    """Calculates the mode."""
    return self.mean()

  def median(self) -> Array:
    """Calculates the median."""
    return self.mean()

  def __getitem__(self, index) -> 'Logistic':
    """See `Distribution.__getitem__`."""
    index = distribution.to_batch_shape_index(self.batch_shape, index)
    return Logistic(loc=self.loc[index], scale=self.scale[index])



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""GumbelCDF bijector."""

from typing import Tuple

from distrax._src.bijectors import bijector as base
import jax.numpy as jnp

Array = base.Array


class GumbelCDF(base.Bijector):
  """A bijector that computes the Gumbel cumulative density function (CDF).

  The Gumbel CDF is given by `y = f(x) = exp(-exp(-x))` for a scalar input `x`.
  Its inverse is `x = -log(-log(y))`. The log-det Jacobian of the transformation
  is `log df/dx = -exp(-x) - x`.
  """

  def __init__(self):
    """Initializes a GumbelCDF bijector."""
    super().__init__(event_ndims_in=0)

  def forward_and_log_det(self, x: Array) -> Tuple[Array, Array]:
    """Computes y = f(x) and log|det J(f)(x)|."""
    exp_neg_x = jnp.exp(-x)
    y = jnp.exp(-exp_neg_x)
    log_det = - x - exp_neg_x
    return y, log_det

  def inverse_and_log_det(self, y: Array) -> Tuple[Array, Array]:
    """Computes x = f^{-1}(y) and log|det J(f^{-1})(y)|."""
    log_y = jnp.log(y)
    x = -jnp.log(-log_y)
    return x, x - log_y

  def same_as(self, other: base.Bijector) -> bool:
    """Returns True if this bijector is guaranteed to be the same as `other`."""
    return type(other) is GumbelCDF  # pylint: disable=unidiomatic-typecheck
